# -*- coding: utf-8 -*-

# Copyright (C) 2010-2013 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
X2GoControlSessionSTDOUT class - core functions for handling your individual X2Go sessions.

This backend handles X2Go server implementations that respond via server-side STDOUT.

"""
__NAME__ = 'x2gocontrolsession-pylib'

# modules
import os
import types
import paramiko
import gevent
import copy
import string
import random
import re
import locale
import threading
import cStringIO

from gevent import socket

# Python X2Go modules
import x2go.sshproxy as sshproxy
import x2go.log as log
import x2go.utils as utils
import x2go.x2go_exceptions as x2go_exceptions
import x2go.defaults as defaults
import x2go.checkhosts as checkhosts

from x2go.backends.terminal import X2GoTerminalSession as _X2GoTerminalSession
from x2go.backends.info import X2GoServerSessionInfo as _X2GoServerSessionInfo
from x2go.backends.info import X2GoServerSessionList as _X2GoServerSessionList
from x2go.backends.proxy import X2GoProxy as _X2GoProxy

import x2go._paramiko
x2go._paramiko.monkey_patch_paramiko()

def _rerewrite_blanks(cmd):
    """\
    In command strings X2Go server scripts expect blanks being rewritten to ,,X2GO_SPACE_CHAR''.
    Commands get rewritten in the terminal sessions. This re-rewrite function helps
    displaying command string in log output.

    @param cmd: command that has to be rewritten for log output
    @type cmd: C{str}

    @return: the command with ,,X2GO_SPACE_CHAR'' re-replaced by blanks
    @rtype: C{str}

    """
    # X2Go run command replace X2GO_SPACE_CHAR string with blanks
    if cmd:
        cmd = cmd.replace("X2GO_SPACE_CHAR", " ")
    return cmd

def _rewrite_password(cmd, user=None, password=None):
    """\
    In command strings Python X2Go replaces some macros with actual values:

      - X2GO_USER -> the user name under which the user is authenticated via SSH
      - X2GO_PASSWORD -> the password being used for SSH authentication

    Both macros can be used to on-the-fly authenticate via RDP.

    @param cmd: command that is to be sent to an X2Go server script
    @type cmd: C{str}
    @param user: the SSH authenticated user name
    @type password: the password being used for SSH authentication

    @return: the command with macros replaced
    @rtype: C{str}

    """
    # if there is a ,,-u X2GO_USER'' parameter in RDP options then we will replace 
    # it by our X2Go session password
    if cmd and user:
        cmd = cmd.replace('X2GO_USER', user)
    # if there is a ,,-p X2GO_PASSWORD'' parameter in RDP options then we will replace 
    # it by our X2Go session password
    if cmd and password:
        cmd = cmd.replace('X2GO_PASSWORD', password)
    return cmd


class X2GoControlSessionSTDOUT(paramiko.SSHClient):
    """\
    In the Python X2Go concept, X2Go sessions fall into two parts: a control session and one to many terminal sessions.

    The control session handles the SSH based communication between server and client. It is mainly derived from
    C{paramiko.SSHClient} and adds on X2Go related functionality.

    """
    associated_terminals = None

    def __init__(self,
                 profile_name='UNKNOWN',
                 add_to_known_hosts=False,
                 known_hosts=None,
                 forward_sshagent=False,
                 unique_hostkey_aliases=False,
                 terminal_backend=_X2GoTerminalSession,
                 info_backend=_X2GoServerSessionInfo,
                 list_backend=_X2GoServerSessionList,
                 proxy_backend=_X2GoProxy,
                 client_rootdir=os.path.join(defaults.LOCAL_HOME, defaults.X2GO_CLIENT_ROOTDIR),
                 sessions_rootdir=os.path.join(defaults.LOCAL_HOME, defaults.X2GO_SESSIONS_ROOTDIR),
                 ssh_rootdir=os.path.join(defaults.LOCAL_HOME, defaults.X2GO_SSH_ROOTDIR),
                 logger=None, loglevel=log.loglevel_DEFAULT,
                 published_applications_no_submenus=0,
                 low_latency=False,
                 **kwargs):
        """\
        Initialize an X2Go control session. For each connected session profile there will be one SSH-based
        control session and one to many terminal sessions that all server-client-communicate via this one common control
        session.

        A control session normally gets set up by an L{X2GoSession} instance. Do not use it directly!!!

        @param profile_name: the profile name of the session profile this control session works for
        @type profile_name: C{str}
        @param add_to_known_hosts: Auto-accept server host validity?
        @type add_to_known_hosts: C{bool}
        @param known_hosts: the underlying Paramiko/SSH systems C{known_hosts} file
        @type known_hosts: C{str}
        @param forward_sshagent: forward SSH agent authentication requests to the X2Go client-side
        @type forward_sshagent: C{bool}
        @param unique_hostkey_aliases: instead of storing [<hostname>]:<port> in known_hosts file, use the
            (unique-by-design) profile ID
        @type unique_hostkey_aliases: C{bool}
        @param terminal_backend: X2Go terminal session backend to use
        @type terminal_backend: C{class}
        @param info_backend: backend for handling storage of server session information
        @type info_backend: C{X2GoServerSessionInfo*} instance
        @param list_backend: backend for handling storage of session list information
        @type list_backend: C{X2GoServerSessionList*} instance
        @param proxy_backend: backend for handling the X-proxy connections
        @type proxy_backend: C{X2GoProxy*} instance
        @param client_rootdir: client base dir (default: ~/.x2goclient)
        @type client_rootdir: C{str}
        @param sessions_rootdir: sessions base dir (default: ~/.x2go)
        @type sessions_rootdir: C{str}
        @param ssh_rootdir: ssh base dir (default: ~/.ssh)
        @type ssh_rootdir: C{str}
        @param published_applications_no_submenus: published applications menus with less items than C{published_applications_no_submenus}
            are rendered without submenus
        @type published_applications_no_submenus: C{int}
        @param logger: you can pass an L{X2GoLogger} object to the
            L{X2GoControlSessionSTDOUT} constructor
        @type logger: L{X2GoLogger} instance
        @param loglevel: if no L{X2GoLogger} object has been supplied a new one will be
            constructed with the given loglevel
        @type loglevel: C{int}
        @param low_latency: set this boolean switch for weak connections, it will double all timeout values.
        @type low_latency: C{bool}
        @param kwargs: catch any non-defined parameters in C{kwargs}
        @type kwargs: C{dict}

        """
        self.associated_terminals = {}
        self.terminated_terminals = []

        self.profile_name = profile_name
        self.add_to_known_hosts = add_to_known_hosts
        self.known_hosts = known_hosts
        self.forward_sshagent = forward_sshagent
        self.unique_hostkey_aliases = unique_hostkey_aliases

        self.hostname = None
        self.port = None

        self.sshproxy_session = None

        self._session_auth_rsakey = None
        self._remote_home = None
        self._remote_group = {}
        self._remote_username = None
        self._remote_peername = None

        self._server_versions = None
        self._server_features = None

        if logger is None:
            self.logger = log.X2GoLogger(loglevel=loglevel)
        else:
            self.logger = copy.deepcopy(logger)
        self.logger.tag = __NAME__

        self._terminal_backend = terminal_backend
        self._info_backend = info_backend
        self._list_backend = list_backend
        self._proxy_backend = proxy_backend

        self.client_rootdir = client_rootdir
        self.sessions_rootdir = sessions_rootdir
        self.ssh_rootdir = ssh_rootdir

        self._published_applications_menu = {}

        self.agent_chan = None
        self.agent_handler = None

        paramiko.SSHClient.__init__(self)
        if self.add_to_known_hosts:
            self.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        self.session_died = False

        self.low_latency = low_latency

        self.published_applications_no_submenus = published_applications_no_submenus
        self._already_querying_published_applications = threading.Lock()

        self._transport_lock = threading.Lock()

    def get_hostname(self):
        """\
        Get the hostname as stored in the properties of this control session.

        @return: the hostname of the connected X2Go server
        @rtype: C{str}

        """
        return self.hostname

    def get_port(self):
        """\
        Get the port number of the SSH connection as stored in the properties of this control session.

        @return: the server-side port number of the control session's SSH connection
        @rtype: C{str}

        """
        return self.port

    def load_session_host_keys(self):
        """\
        Load known SSH host keys from the C{known_hosts} file.

        If the file does not exist, create it first.

        """
        if self.known_hosts is not None:
            utils.touch_file(self.known_hosts)
            self.load_host_keys(self.known_hosts)

    def __del__(self):
        """\
        On instance descruction, do a proper session disconnect from the server.

        """
        self.disconnect()

    def _x2go_sftp_put(self, local_path, remote_path):
        """
        Put a local file on the remote server via sFTP.

        During sFTP operations, remote command execution gets blocked.

        @param local_path: full local path name of the file to be put on the server
        @type local_path: C{str}
        @param remote_path: full remote path name of the server-side target location, path names have to be Unix-compliant
        @type remote_path: C{str}

        @raise X2GoControlSessionException: if the SSH connection dropped out

        """
        ssh_transport = self.get_transport()
        self._transport_lock.acquire()
        if ssh_transport and ssh_transport.is_authenticated():
            self.logger('sFTP-put: %s -> %s:%s' % (os.path.normpath(local_path), self.remote_peername(), remote_path), loglevel=log.loglevel_DEBUG)
            self.sftp_client = paramiko.SFTPClient.from_transport(ssh_transport)
            try:
                self.sftp_client.put(os.path.normpath(local_path), remote_path)
            except (x2go_exceptions.SSHException, socket.error, IOError):
                # react to connection dropped error for SSH connections
                self.session_died = True
                self._transport_lock.release()
                raise x2go_exceptions.X2GoControlSessionException('The SSH connection was dropped during an sFTP put action.')
            self.sftp_client = None
        self._transport_lock.release()

    def _x2go_sftp_write(self, remote_path, content):
        """
        Create a text file on the remote server via sFTP.

        During sFTP operations, remote command execution gets blocked.

        @param remote_path: full remote path name of the server-side target location, path names have to be Unix-compliant
        @type remote_path: C{str}
        @param content: a text file, multi-line files use Unix-link EOL style
        @type content: C{str}

        @raise X2GoControlSessionException: if the SSH connection dropped out

        """
        ssh_transport = self.get_transport()
        self._transport_lock.acquire()
        if ssh_transport and ssh_transport.is_authenticated():
            self.logger('sFTP-write: opening remote file %s on host %s for writing' % (remote_path, self.remote_peername()), loglevel=log.loglevel_DEBUG)
            self.sftp_client = paramiko.SFTPClient.from_transport(ssh_transport)
            try:
                remote_fileobj = self.sftp_client.open(remote_path, 'w')
                self.logger('sFTP-write: writing content: %s' % content, loglevel=log.loglevel_DEBUG_SFTPXFER)
                remote_fileobj.write(content)
                remote_fileobj.close()
            except (x2go_exceptions.SSHException, socket.error, IOError):
                self.session_died = True
                self._transport_lock.release()
                self.logger('sFTP-write: opening remote file %s on host %s failed' % (remote_path, self.remote_peername()), loglevel=log.loglevel_WARN)
                raise x2go_exceptions.X2GoControlSessionException('The SSH connection was dropped during an sFTP write action.')
            self.sftp_client = None
        self._transport_lock.release()

    def _x2go_sftp_remove(self, remote_path):
        """
        Remote a remote file from the server via sFTP.

        During sFTP operations, remote command execution gets blocked.

        @param remote_path: full remote path name of the server-side file to be removed, path names have to be Unix-compliant
        @type remote_path: C{str}

        @raise X2GoControlSessionException: if the SSH connection dropped out

        """
        ssh_transport = self.get_transport()
        self._transport_lock.acquire()
        if ssh_transport and ssh_transport.is_authenticated():
            self.logger('sFTP-write: removing remote file %s on host %s' % (remote_path, self.remote_peername()), loglevel=log.loglevel_DEBUG)
            self.sftp_client = paramiko.SFTPClient.from_transport(ssh_transport)
            try:
                self.sftp_client.remove(remote_path)
            except (x2go_exceptions.SSHException, socket.error, IOError):
                self.session_died = True
                self._transport_lock.release()
                self.logger('sFTP-write: removing remote file %s on host %s failed' % (remote_path, self.remote_peername()), loglevel=log.loglevel_WARN)
                raise x2go_exceptions.X2GoControlSessionException('The SSH connection was dropped during an sFTP remove action.')
            self.sftp_client = None
            self._transport_lock.release()

    def _x2go_exec_command(self, cmd_line, loglevel=log.loglevel_INFO, timeout=20, **kwargs):
        """
        Execute an X2Go server-side command via SSH.

        During SSH command executions, sFTP operations get blocked.

        @param cmd_line: the command to be executed on the remote server
        @type cmd_line: C{str} or C{list}
        @param loglevel: use this loglevel for reporting about remote command execution
        @type loglevel: C{int}
        @param timeout: if commands take longer than C{<timeout>} to be executed, consider the control session connection
            to have died.
        @type timeout: C{int}
        @param kwargs: parameters that get passed through to the C{paramiko.SSHClient.exec_command()} method.
        @type kwargs: C{dict}

        @return: C{True} if the command could be successfully executed on the remote X2Go server
        @rtype: C{bool}

        @raise X2GoControlSessionException: if the command execution failed (due to a lost connection)

        """
        if type(cmd_line) == types.ListType:
            cmd = " ".join(cmd_line)
        else:
            cmd = cmd_line

        cmd = 'sh -c \"%s\"' % cmd

        if self.session_died:
            self.logger("control session seams to be dead, not executing command ,,%s'' on X2Go server %s" % (_rerewrite_blanks(cmd), self.profile_name,), loglevel=loglevel)
            return (cStringIO.StringIO(), cStringIO.StringIO(), cStringIO.StringIO('failed to execute command'))

        self._transport_lock.acquire()

        _retval = None

        ssh_transport = self.get_transport()
        if ssh_transport and ssh_transport.is_authenticated():

            if self.low_latency: timeout = timeout * 2
            timer = gevent.Timeout(timeout)
            timer.start()
            try:
                self.logger("executing command on X2Go server ,,%s'': %s" % (self.profile_name, _rerewrite_blanks(cmd)), loglevel=loglevel)
                _retval = self.exec_command(_rewrite_password(cmd, user=self.get_transport().get_username(), password=self._session_password), **kwargs)
            except AttributeError:
                self.session_died = True
                self._transport_lock.release()
                if self.sshproxy_session:
                    self.sshproxy_session.stop_thread()
                raise x2go_exceptions.X2GoControlSessionException('the X2Go control session has died unexpectedly')
            except EOFError:
                self.session_died = True
                self._transport_lock.release()
                if self.sshproxy_session:
                    self.sshproxy_session.stop_thread()
                raise x2go_exceptions.X2GoControlSessionException('the X2Go control session has died unexpectedly')
            except x2go_exceptions.SSHException:
                self.session_died = True
                self._transport_lock.release()
                if self.sshproxy_session:
                    self.sshproxy_session.stop_thread()
                raise x2go_exceptions.X2GoControlSessionException('the X2Go control session has died unexpectedly')
            except gevent.timeout.Timeout:
                self.session_died = True
                self._transport_lock.release()
                if self.sshproxy_session:
                    self.sshproxy_session.stop_thread()
                raise x2go_exceptions.X2GoControlSessionException('the X2Go control session command timed out')
            except socket.error:
                self.session_died = True
                self._transport_lock.release()
                if self.sshproxy_session:
                    self.sshproxy_session.stop_thread()
                raise x2go_exceptions.X2GoControlSessionException('the X2Go control session has died unexpectedly')
            finally:
                timer.cancel()

        else:
            self._transport_lock.release()
            raise x2go_exceptions.X2GoControlSessionException('the X2Go control session is not connected')

        self._transport_lock.release()
        return _retval

    @property
    def _x2go_server_versions(self):
        """\
        Render a dictionary of server-side X2Go components and their versions. Results get cached
        once there has been one successful query.

        """
        if self._server_versions is None:
            self._server_versions = {}
            (stdin, stdout, stderr) = self._x2go_exec_command('which x2goversion >/dev/null && x2goversion')
            _lines = stdout.read().split('\n')
            for _line in _lines:
                if ':' not in _line: continue
                comp = _line.split(':')[0].strip()
                version = _line.split(':')[1].strip()
                self._server_versions.update({comp: version})
            self.logger('server-side X2Go components and their versions are: %s' % self._server_versions, loglevel=log.loglevel_DEBUG)
        return self._server_versions

    def query_server_versions(self, force=False):
        """\
        Do a query for the server-side list of X2Go components and their versions.

        @param force: do not use the cached component list, really ask the server (again)
        @type force: C{bool}

        @return: dictionary of X2Go components (as keys) and their versions (as values)
        @rtype: C{list}

        """
        if force:
            self._server_versions = None
        return self._x2go_server_versions
    get_server_versions = query_server_versions

    @property
    def _x2go_server_features(self):
        """\
        Render a list of server-side X2Go features. Results get cached once there has been one successful query.

        """
        if self._server_features is None:
            (stdin, stdout, stderr) = self._x2go_exec_command('which x2gofeaturelist >/dev/null && x2gofeaturelist')
            self._server_features = stdout.read().split('\n')
            self._server_features = [ f for f in self._server_features if f ]
            self._server_features.sort()
            self.logger('server-side X2Go features are: %s' % self._server_features, loglevel=log.loglevel_DEBUG)
        return self._server_features

    def query_server_features(self, force=False):
        """\
        Do a query for the server-side list of X2Go features.

        @param force: do not use the cached feature list, really ask the server (again)
        @type force: C{bool}

        @return: list of X2Go feature names
        @rtype: C{list}

        """
        if force:
            self._server_features = None
        return self._x2go_server_features
    get_server_features = query_server_features

    @property
    def _x2go_remote_home(self):
        """\
        Retrieve and cache the remote home directory location.

        """
        if self._remote_home is None:
            (stdin, stdout, stderr) = self._x2go_exec_command('echo $HOME')
            stdout_r = stdout.read()
            if stdout_r:
                self._remote_home = stdout_r.split()[0]
                self.logger('remote user\' home directory: %s' % self._remote_home, loglevel=log.loglevel_DEBUG)
            return self._remote_home
        else:
            return self._remote_home

    def _x2go_remote_group(self, group):
        """\
        Retrieve and cache the members of a server-side POSIX group.

        @param group: remote POSIX group name
        @type group: C{str}

        @return: list of POSIX group members
        @rtype: C{list}

        """
        if not self._remote_group.has_key(group):
            (stdin, stdout, stderr) = self._x2go_exec_command('getent group %s | cut -d":" -f4' % group)
            self._remote_group[group] = stdout.read().split('\n')[0].split(',')
            self.logger('remote %s group: %s' % (group, self._remote_group[group]), loglevel=log.loglevel_DEBUG)
            return self._remote_group[group]
        else:
            return self._remote_group[group]

    def is_x2gouser(self, username):
        """\
        Is the remote user allowed to launch X2Go sessions?

        FIXME: this method is currently non-functional.

        @param username: remote user name
        @type username: C{str}

        @return: C{True} if the remote user is allowed to launch X2Go sessions
        @rtype: C{bool}

        """
        ###
        ### FIXME:
        ###
        # discussion about server-side access restriction based on posix group membership or similar currently 
        # in process (as of 20110517, mg)
        #return username in self._x2go_remote_group('x2gousers')
        return True

    def is_sshfs_available(self):
        """\
        Check if the remote user is allowed to use SSHFS mounts.

        @return: C{True} if the user is allowed to connect client-side shares to the X2Go session
        @rtype: C{bool}

        """
        if self.remote_username() in self._x2go_remote_group('fuse'):
            return True
        return False

    def remote_username(self):
        """\
        Returns (and caches) the control session's remote username.

        @return: SSH transport's user name
        @rtype: C{str}

        @raise X2GoControlSessionException: on SSH connection loss

        """
        if self._remote_username is None:
            if self.get_transport() is not None:
                try:
                    self._remote_username = self.get_transport().get_username()
                except:
                    self.session_died = True
                    raise x2go_exceptions.X2GoControlSessionException('Lost connection to X2Go server')
        return self._remote_username

    def remote_peername(self):
        """\
        Returns (and caches) the control session's remote host (name or ip).

        @return: SSH transport's peer name
        @rtype: C{tuple}

        @raise X2GoControlSessionException: on SSH connection loss

        """
        if self._remote_peername is None:
            if self.get_transport() is not None:
                try:
                    self._remote_peername = self.get_transport().getpeername()
                except:
                    self.session_died = True
                    raise x2go_exceptions.X2GoControlSessionException('Lost connection to X2Go server')
        return self._remote_peername

    @property
    def _x2go_session_auth_rsakey(self):
        """\
        Generate (and cache) a temporary RSA host key for the lifetime of this control session.

        """
        if self._session_auth_rsakey is None:
            self._session_auth_rsakey = paramiko.RSAKey.generate(defaults.RSAKEY_STRENGTH)
        return self._session_auth_rsakey

    def set_profile_name(self, profile_name):
        """\
        Manipulate the control session's profile name.

        @param profile_name: new profile name for this control session
        @type profile_name: C{str}

        """
        self.profile_name = profile_name

    def check_host(self, hostname, port=22):
        """\
        Wraps around a Paramiko/SSH host key check.

        @param hostname: the remote X2Go server's hostname
        @type hostname: C{str}
        @param port: the SSH port of the remote X2Go server
        @type port: C{int}

        @return: C{True} if the host key check succeeded, C{False} otherwise
        @rtype: C{bool}

        """
        # trailing whitespace tolerance
        hostname = hostname.strip()

        # force into IPv4 for localhost connections
        if hostname in ('localhost', 'localhost.localdomain'):
            hostname = '127.0.0.1'

        return checkhosts.check_ssh_host_key(self, hostname, port=port)

    def connect(self, hostname, port=22, username='', password='', pkey=None,
                key_filename=None, timeout=None, allow_agent=False, look_for_keys=False,
                use_sshproxy=False, sshproxy_host='', sshproxy_port=22, sshproxy_user='', sshproxy_password='', sshproxy_force_password_auth=False,
                sshproxy_key_filename='', sshproxy_pkey=None, sshproxy_look_for_keys=False, sshproxy_allow_agent=False,
                sshproxy_tunnel='',
                forward_sshagent=None,
                unique_hostkey_aliases=None,
                session_instance=None,
                add_to_known_hosts=False, force_password_auth=False):
        """\
        Connect to an X2Go server and authenticate to it. This method is directly
        inherited from the C{paramiko.SSHClient} class. The features of the Paramiko
        SSH client connect method are recited here. The parameters C{add_to_known_hosts},
        C{force_password_auth}, C{session_instance} and all SSH proxy related parameters
        have been added as X2Go specific parameters

        The server's host key is checked against the system host keys 
        (see C{load_system_host_keys}) and any local host keys (C{load_host_keys}).
        If the server's hostname is not found in either set of host keys, the missing host
        key policy is used (see C{set_missing_host_key_policy}).  The default policy is
        to reject the key and raise an C{SSHException}.

        Authentication is attempted in the following order of priority:

            - The C{pkey} or C{key_filename} passed in (if any)
            - Any key we can find through an SSH agent
            - Any "id_rsa" or "id_dsa" key discoverable in C{~/.ssh/}
            - Plain username/password auth, if a password was given

        If a private key requires a password to unlock it, and a password is
        passed in, that password will be used to attempt to unlock the key.

        @param hostname: the server to connect to
        @type hostname: C{str}
        @param port: the server port to connect to
        @type port: C{int}
        @param username: the username to authenticate as (defaults to the
            current local username)
        @type username: C{str}
        @param password: a password to use for authentication or for unlocking
            a private key
        @type password: C{str}
        @param key_filename: the filename, or list of filenames, of optional
            private key(s) to try for authentication
        @type key_filename: C{str} or list(str)
        @param pkey: an optional private key to use for authentication
        @type pkey: C{PKey}
        @param forward_sshagent: forward SSH agent authentication requests to the X2Go client-side
            (will update the class property of the same name)
        @type forward_sshagent: C{bool}
        @param unique_hostkey_aliases: update the unique_hostkey_aliases class property
        @type unique_hostkey_aliases: C{bool}
        @param timeout: an optional timeout (in seconds) for the TCP connect
        @type timeout: float
        @param look_for_keys: set to C{True} to enable searching for discoverable
            private key files in C{~/.ssh/}
        @type look_for_keys: C{bool}
        @param allow_agent: set to C{True} to enable connecting to a local SSH agent
            for acquiring authentication information
        @type allow_agent: C{bool}
        @param add_to_known_hosts: non-paramiko option, if C{True} paramiko.AutoAddPolicy() 
            is used as missing-host-key-policy. If set to C{False} paramiko.RejectPolicy() 
            is used
        @type add_to_known_hosts: C{bool}
        @param force_password_auth: non-paramiko option, disable pub/priv key authentication 
            completely, even if the C{pkey} or the C{key_filename} parameter is given
        @type force_password_auth: C{bool}
        @param session_instance: an instance L{X2GoSession} using this L{X2GoControlSessionSTDOUT}
            instance.
        @type session_instance: C{obj}
        @param use_sshproxy: connect through an SSH proxy
        @type use_sshproxy: C{True} if an SSH proxy is to be used for tunneling the connection
        @param sshproxy_host: hostname of the SSH proxy server
        @type sshproxy_host: C{str}
        @param sshproxy_port: port of the SSH proxy server
        @type sshproxy_port: C{int}
        @param sshproxy_user: username that we use for authenticating against C{<sshproxy_host>}
        @type sshproxy_user: C{str}
        @param sshproxy_password: a password to use for SSH proxy authentication or for unlocking
            a private key
        @type sshproxy_password: C{str}
        @param sshproxy_force_password_auth: enforce using a given C{sshproxy_password} even if a key(file) is given
        @type sshproxy_force_password_auth: C{bool}
        @param sshproxy_key_filename: local file location of the private key file
        @type sshproxy_key_filename: C{str}
        @param sshproxy_pkey: an optional private key to use for SSH proxy authentication
        @type sshproxy_pkey: C{PKey}
        @param sshproxy_look_for_keys: set to C{True} to enable connecting to a local SSH agent
            for acquiring authentication information (for SSH proxy authentication)
        @type sshproxy_look_for_keys: C{bool}
        @param sshproxy_allow_agent: set to C{True} to enable connecting to a local SSH agent
            for acquiring authentication information (for SSH proxy authentication)
        @type sshproxy_allow_agent: C{bool}
        @param sshproxy_tunnel: the SSH proxy tunneling parameters, format is: <local-address>:<local-port>:<remote-address>:<remote-port>
        @type sshproxy_tunnel: C{str}

        @return: C{True} if an authenticated SSH transport could be retrieved by this method
        @rtype: C{bool}

        @raise BadHostKeyException: if the server's host key could not be
            verified
        @raise AuthenticationException: if authentication failed
        @raise SSHException: if there was any other error connecting or
            establishing an SSH session
        @raise socket.error: if a socket error occurred while connecting
        @raise X2GoSSHProxyException: any SSH proxy exception is passed through while establishing the SSH proxy connection and tunneling setup
        @raise X2GoSSHAuthenticationException: any SSH proxy authentication exception is passed through while establishing the SSH proxy connection and tunneling setup
        @raise X2GoRemoteHomeException: if the remote home directory does not exist or is not accessible

        """
        _fake_hostname = None

        if unique_hostkey_aliases is not None:
            self.unique_hostkey_aliases = unique_hostkey_aliases
            # prep the fake hostname with the real hostname, so we trigger the corresponding code path in 
            # x2go.checkhosts and either of its missing host key policies
            if self.unique_hostkey_aliases:
                _fake_hostname = "[%s]:%s" % (hostname, port)

        if use_sshproxy and sshproxy_host and sshproxy_user:
            try:

                if not sshproxy_tunnel:
                    sshproxy_tunnel = "localhost:44444:%s:%s" % (hostname, port)
                self.sshproxy_session = sshproxy.X2GoSSHProxy(known_hosts=self.known_hosts,
                                                              sshproxy_host=sshproxy_host,
                                                              sshproxy_port=sshproxy_port,
                                                              sshproxy_user=sshproxy_user,
                                                              sshproxy_password=sshproxy_password,
                                                              sshproxy_force_password_auth=sshproxy_force_password_auth,
                                                              sshproxy_key_filename=sshproxy_key_filename,
                                                              sshproxy_pkey=sshproxy_pkey,
                                                              sshproxy_look_for_keys=sshproxy_look_for_keys,
                                                              sshproxy_allow_agent=sshproxy_allow_agent,
                                                              sshproxy_tunnel=sshproxy_tunnel,
                                                              session_instance=session_instance,
                                                              logger=self.logger,
                                                             )
                hostname = self.sshproxy_session.get_local_proxy_host()
                port = self.sshproxy_session.get_local_proxy_port()
                _fake_hostname = self.sshproxy_session.get_remote_host()
                _fake_port = self.sshproxy_session.get_remote_port()
                _fake_hostname = "[%s]:%s" % (_fake_hostname, _fake_port)

            except:
                if self.sshproxy_session:
                    self.sshproxy_session.stop_thread()
                self.sshproxy_session = None
                raise

            if self.sshproxy_session is not None:
                self.sshproxy_session.start()

                # divert port to sshproxy_session's local forwarding port (it might have changed due to 
                # SSH connection errors
                gevent.sleep(.1)
                port = self.sshproxy_session.get_local_proxy_port()

        if not add_to_known_hosts and session_instance:
            self.set_missing_host_key_policy(checkhosts.X2GoInteractiveAddPolicy(caller=self, session_instance=session_instance, fake_hostname=_fake_hostname))

        if add_to_known_hosts:
            self.set_missing_host_key_policy(checkhosts.X2GoAutoAddPolicy(caller=self, session_instance=session_instance, fake_hostname=_fake_hostname))

        # disable pub/priv key authentication if forced
        if force_password_auth:
            key_filename = None
            pkey = None

        # trailing whitespace tolerance in hostname
        hostname = hostname.strip()

        self.logger('connecting to [%s]:%s' % (hostname, port), loglevel=log.loglevel_NOTICE)

        self.load_session_host_keys()

        _hostname = hostname
        # enforce IPv4 for localhost address
        if _hostname in ('localhost', 'localhost.localdomain'):
            _hostname = '127.0.0.1'

        # update self.forward_sshagent via connect method parameter
        if forward_sshagent is not None:
            self.forward_sshagent = forward_sshagent

        if timeout and self.low_latency:
            timeout = timeout * 2

        if key_filename or pkey or look_for_keys or allow_agent or (password and force_password_auth):
            try:
                if password and force_password_auth:
                    self.logger('trying keyboard-interactive SSH authentication with server', loglevel=log.loglevel_DEBUG)
                    paramiko.SSHClient.connect(self, _hostname, port=port, username=username, pkey=None, password=password,
                                               key_filename=None, timeout=timeout, allow_agent=False,
                                               look_for_keys=False)
                elif (key_filename and os.path.exists(os.path.normpath(key_filename))) or pkey:
                    self.logger('trying SSH pub/priv key authentication with server', loglevel=log.loglevel_DEBUG)
                    paramiko.SSHClient.connect(self, _hostname, port=port, username=username, pkey=pkey,
                                               key_filename=key_filename, timeout=timeout, allow_agent=allow_agent,
                                               look_for_keys=look_for_keys)
                else:
                    self.logger('trying SSH key discovery or agent authentication with server', loglevel=log.loglevel_DEBUG)
                    try:
                        paramiko.SSHClient.connect(self, _hostname, port=port, username=username, pkey=None,
                                                      key_filename=None, timeout=timeout, allow_agent=allow_agent,
                                                      look_for_keys=look_for_keys)
                    except paramiko.SSHException, e:
                        if str(e) == 'No authentication methods available':
                            raise paramiko.AuthenticationException('Interactive password authentication required!')
                        else:
                            self.close()
                            if self.sshproxy_session:
                                self.sshproxy_session.stop_thread()
                            raise(e)

                # since Paramiko 1.7.7.1 there is compression available, let's use it if present...
                t = self.get_transport()
                if x2go._paramiko.PARAMIKO_FEATURE['use-compression']:
                    t.use_compression(compress=True)
                # enable keep alive callbacks
                t.set_keepalive(5)

            except paramiko.AuthenticationException, e:
                self.close()
                if password:
                    self.logger('next auth mechanism we\'ll try is keyboard-interactive authentication', loglevel=log.loglevel_DEBUG)
                    try:
                        paramiko.SSHClient.connect(self, _hostname, port=port, username=username, password=password,
                                                   key_filename=None, pkey=None, timeout=timeout, allow_agent=False, look_for_keys=False)
                    except:
                        self.close()
                        if self.sshproxy_session:
                            self.sshproxy_session.stop_thread()
                        raise
                else:
                    self.close()
                    if self.sshproxy_session:
                        self.sshproxy_session.stop_thread()
                    raise(e)

            except:
                self.close()
                if self.sshproxy_session:
                    self.sshproxy_session.stop_thread()
                raise

        # if there is no private key (and no agent auth), we will use the given password, if any
        else:
            # create a random password if password is empty to trigger host key validity check
            if not password:
                password = "".join([random.choice(string.letters+string.digits) for x in range(1, 20)])
            self.logger('performing SSH keyboard-interactive authentication with server', loglevel=log.loglevel_DEBUG)
            try:
                paramiko.SSHClient.connect(self, _hostname, port=port, username=username, password=password,
                                           timeout=timeout, allow_agent=False, look_for_keys=False)
            except paramiko.AuthenticationException, e:
                self.close()
                if self.sshproxy_session:
                    self.sshproxy_session.stop_thread()
                raise e
            except:
                self.close()
                if self.sshproxy_session:
                    self.sshproxy_session.stop_thread()
                raise

        self.set_missing_host_key_policy(paramiko.RejectPolicy())

        self.hostname = hostname
        self.port = port

        # preparing reverse tunnels
        ssh_transport = self.get_transport()
        ssh_transport.reverse_tunnels = {}

        # mark Paramiko/SSH transport as X2GoControlSession
        ssh_transport._x2go_session_marker = True
        self._session_password = password

        if ssh_transport is not None:
            self.session_died = False
            self.query_server_features(force=True)
            if self.forward_sshagent:
                if x2go._paramiko.PARAMIKO_FEATURE['forward-ssh-agent']:
                    self.agent_chan = ssh_transport.open_session()
                    self.agent_handler = paramiko.agent.AgentRequestHandler(self.agent_chan)
                    self.logger('Requesting SSH agent forwarding for control session of connected session profile %s' % self.profile_name, loglevel=log.loglevel_INFO)
                else:
                    self.logger('SSH agent forwarding is not available in the Paramiko version used with this instance of Python X2Go', loglevel=log.loglevel_WARN)
        else:
            self.close()
            if self.sshproxy_session:
                self.sshproxy_session.stop_thread()

        self._remote_home = None
        if not self.home_exists():
            self.close()
            if self.sshproxy_session:
                self.sshproxy_session.stop_thread()
            raise x2go_exceptions.X2GoRemoteHomeException('remote home directory does not exist')

        return (self.get_transport() is not None)

    def dissociate(self, terminal_session):
        """\
        Drop an associated terminal session.

        @param terminal_session: the terminal session object to remove from the list of associated terminals
        @type terminal_session: C{X2GoTerminalSession*}

        """
        for t_name in self.associated_terminals.keys():
            if self.associated_terminals[t_name] == terminal_session:
                del self.associated_terminals[t_name]
                if self.terminated_terminals.has_key(t_name):
                    del self.terminated_terminals[t_name]

    def disconnect(self):
        """\
        Disconnect this control session from the remote server.

        @return: report success or failure after having disconnected
        @rtype: C{bool}

        """
        if self.associated_terminals:
            t_names = self.associated_terminals.keys()
            for  t_obj in self.associated_terminals.values():
                try:
                    if not self.session_died:
                        t_obj.suspend()
                except x2go_exceptions.X2GoTerminalSessionException:
                    pass
                except x2go_exceptions.X2GoControlSessionException:
                    self.session_died
                t_obj.__del__()
            for t_name in t_names:
                try:
                    del self.associated_terminals[t_name]
                except KeyError:
                    pass

        self._remote_home = None
        self._remote_group = {}

        self._session_auth_rsakey = None

        # in any case, release out internal transport lock
        self._transport_lock.release()

        # close SSH agent auth forwarding objects
        if self.agent_handler is not None:
            self.agent_handler.close()

        if self.agent_chan is not None:
            self.agent_chan.close()

        retval = False
        try:
            if self.get_transport() is not None:
                retval = self.get_transport().is_active()
                try:
                    self.close()
                except IOError:
                    pass
        except AttributeError:
            # if the Paramiko _transport object has not yet been initialized, ignore it
            # but state that this method call did not close the SSH client, but was already closed
            pass

        # take down sshproxy_session no matter what happened to the control session itself
        if self.sshproxy_session is not None:
            self.sshproxy_session.stop_thread()

        return retval

    def home_exists(self):
        """\
        Test if the remote home directory exists.

        @return: C{True} if the home directory exists, C{False} otherwise
        @rtype: C{bool}

        """
        (_stdin, _stdout, _stderr) = self._x2go_exec_command('stat -tL "%s"' % self._x2go_remote_home, loglevel=log.loglevel_DEBUG)
        if _stdout.read():
            return True
        return False


    def is_alive(self):
        """\
        Test if the connection to the remote X2Go server is still alive.

        @return: C{True} if the connection is still alive, C{False} otherwise
        @rtype: C{bool}

        """
        try:
            if self._x2go_exec_command('echo', loglevel=log.loglevel_DEBUG):
                return True
        except x2go_exceptions.X2GoControlSessionException:
            self.session_died = True
            self.disconnect()
        return False

    def has_session_died(self):
        """\
        Test if the connection to the remote X2Go server died on the way.

        @return: C{True} if the connection has died, C{False} otherwise
        @rtype: C{bool}

        """
        return self.session_died

    def get_published_applications(self, lang=None, refresh=False, raw=False, very_raw=False, max_no_submenus=defaults.PUBAPP_MAX_NO_SUBMENUS):
        """\
        Retrieve the menu tree of published applications from the remote X2Go server.

        The C{raw} option lets this method return a C{list} of C{dict} elements. Each C{dict} elements has a 
        C{desktop} key containing a shortened version of the text output of a .desktop file and an C{icon} key
        which contains the desktop base64-encoded icon data.

        The {very_raw} lets this method return the output of the C{x2gogetapps} script as is.

        @param lang: locale/language identifier
        @type lang: C{str}
        @param refresh: force reload of the menu tree from X2Go server
        @type refresh: C{bool}
        @param raw: retrieve a raw output of the server list of published applications
        @type raw: C{bool}
        @param very_raw: retrieve a very raw output of the server list of published applications
        @type very_raw: C{bool}

        @return: an i18n capable menu tree packed as a Python dictionary
        @rtype: C{list}

        """
        self._already_querying_published_applications.acquire()

        if defaults.X2GOCLIENT_OS != 'Windows' and lang is None:
            lang = locale.getdefaultlocale()[0]
        elif lang is None:
            lang = 'en'

        if 'X2GO_PUBLISHED_APPLICATIONS' in self.get_server_features():
            if self._published_applications_menu is {} or \
               not self._published_applications_menu.has_key(lang) or \
               raw or very_raw or refresh or \
               (self.published_applications_no_submenus != max_no_submenus):

                self.published_applications_no_submenus = max_no_submenus

                ### STAGE 1: retrieve menu from server

                self.logger('querying server (%s) for list of published applications' % self.profile_name, loglevel=log.loglevel_NOTICE)
                (stdin, stdout, stderr) = self._x2go_exec_command('which x2gogetapps >/dev/null && x2gogetapps')
                _raw_output = stdout.read()

                if very_raw:
                    self.logger('published applications query for %s finished, return very raw output' % self.profile_name, loglevel=log.loglevel_NOTICE)
                    self._already_querying_published_applications.release()
                    return _raw_output

                ### STAGE 2: dissect the text file retrieved from server, cut into single menu elements

                _raw_menu_items = _raw_output.split('</desktop>\n')
                _raw_menu_items = [ i.replace('<desktop>\n', '') for i in _raw_menu_items ]
                _menu = []
                for _raw_menu_item in _raw_menu_items:
                    if '<icon>\n' in _raw_menu_item and '</icon>' in _raw_menu_item:
                        _menu_item = _raw_menu_item.split('<icon>\n')[0] + _raw_menu_item.split('</icon>\n')[1]
                        _icon_base64 = _raw_menu_item.split('<icon>\n')[1].split('</icon>\n')[0]
                    else:
                        _menu_item = _raw_menu_item
                        _icon_base64 = None
                    if _menu_item:
                        _menu.append({ 'desktop': _menu_item, 'icon': _icon_base64, })
                        _menu_item = None
                        _icon_base64 = None

                if raw:
                    self.logger('published applications query for %s finished, returning raw output' % self.profile_name, loglevel=log.loglevel_NOTICE)
                    self._already_querying_published_applications.release()
                    return _menu

                if len(_menu) > max_no_submenus >= 0:
                    _render_submenus = True
                else:
                    _render_submenus = False

                # STAGE 3: create menu structure in a Python dictionary

                _category_map = {
                    lang: {
                        'Multimedia': [],
                        'Development': [],
                        'Education': [],
                        'Games': [],
                        'Graphics': [],
                        'Internet': [],
                        'Office': [],
                        'System': [],
                        'Utilities': [],
                        'Other Applications': [],
                        'TOP': [],
                    }
                }
                _empty_menus = _category_map[lang].keys()

                for item in _menu:

                    _menu_entry_name = ''
                    _menu_entry_fallback_name = ''
                    _menu_entry_comment = ''
                    _menu_entry_fallback_comment = ''
                    _menu_entry_exec = ''
                    _menu_entry_cat = ''
                    _menu_entry_shell = False

                    lang_regio = lang
                    lang_only = lang_regio.split('_')[0]

                    for line in item['desktop'].split('\n'):
                        if re.match('^Name\[%s\]=.*' % lang_regio, line) or re.match('Name\[%s\]=.*' % lang_only, line):
                            _menu_entry_name = line.split("=")[1].strip()
                        elif re.match('^Name=.*', line):
                            _menu_entry_fallback_name = line.split("=")[1].strip()
                        elif re.match('^Comment\[%s\]=.*' % lang_regio, line) or re.match('Comment\[%s\]=.*' % lang_only, line):
                            _menu_entry_comment = line.split("=")[1].strip()
                        elif re.match('^Comment=.*', line):
                            _menu_entry_fallback_comment = line.split("=")[1].strip()
                        elif re.match('^Exec=.*', line):
                            _menu_entry_exec = line.split("=")[1].strip()
                        elif re.match('^Terminal=.*(t|T)(r|R)(u|U)(e|E).*', line):
                            _menu_entry_shell = True
                        elif re.match('^Categories=.*', line):
                            if 'X2Go-Top' in line:
                                _menu_entry_cat = 'TOP'
                            elif 'Audio' in line or 'Video' in line:
                                _menu_entry_cat = 'Multimedia'
                            elif 'Development' in line:
                                _menu_entry_cat = 'Development'
                            elif 'Education' in line:
                                _menu_entry_cat = 'Education'
                            elif 'Game' in line:
                                _menu_entry_cat = 'Games'
                            elif 'Graphics' in line:
                                _menu_entry_cat = 'Graphics'
                            elif 'Network' in line:
                                _menu_entry_cat = 'Internet'
                            elif 'Office' in line:
                                _menu_entry_cat = 'Office'
                            elif 'Settings' in line:
                                continue
                            elif 'System' in line:
                                _menu_entry_cat = 'System'
                            elif 'Utility' in line:
                                _menu_entry_cat = 'Utilities'
                            else:
                                _menu_entry_cat = 'Other Applications'

                    if not _menu_entry_exec:
                        continue
                    else:
                        # FIXME: strip off any noted options (%f, %F, %u, %U, ...), this can be more intelligent
                        _menu_entry_exec = _menu_entry_exec.replace('%f', '').replace('%F','').replace('%u','').replace('%U','')
                        if _menu_entry_shell:
                            _menu_entry_exec = "x-terminal-emulator -e '%s'" % _menu_entry_exec

                    if not _menu_entry_cat:
                        _menu_entry_cat = 'Other Applications'

                    if not _render_submenus:
                        _menu_entry_cat = 'TOP'

                    if _menu_entry_cat in _empty_menus:
                        _empty_menus.remove(_menu_entry_cat)

                    if not _menu_entry_name: _menu_entry_name = _menu_entry_fallback_name
                    if not _menu_entry_comment: _menu_entry_comment = _menu_entry_fallback_comment
                    if not _menu_entry_comment: _menu_entry_comment = _menu_entry_name

                    _menu_entry_icon = item['icon']

                    _category_map[lang][_menu_entry_cat].append(
                        {
                            'name': _menu_entry_name,
                            'comment': _menu_entry_comment,
                            'exec': _menu_entry_exec,
                            'icon': _menu_entry_icon,
                        }
                    )

                for _cat in _empty_menus:
                    del _category_map[lang][_cat]

                for _cat in _category_map[lang].keys():
                    _sorted = sorted(_category_map[lang][_cat], key=lambda k: k['name'])
                    _category_map[lang][_cat] = _sorted

                self._published_applications_menu.update(_category_map)
                self.logger('published applications query for %s finished, return menu tree' % self.profile_name, loglevel=log.loglevel_NOTICE)

        else:
            # FIXME: ignoring the absence of the published applications feature for now, handle it appropriately later
            pass

        self._already_querying_published_applications.release()
        return self._published_applications_menu

    def start(self, **kwargs):
        """\
        Start a new X2Go session.

        The L{X2GoControlSessionSTDOUT.start()} method accepts any parameter
        that can be passed to any of the C{X2GoTerminalSession} backend class
        constructors.

        @param kwargs: parameters that get passed through to the control session's
            L{resume()} method, only the C{session_name} parameter will get removed
            before pass-through
        @type kwargs: C{dict}

        @return: return value of the cascaded L{resume()} method, denoting the success or failure
            of the session startup
        @rtype: C{bool}

        """
        if 'session_name' in kwargs.keys():
            del kwargs['session_name']
        return self.resume(**kwargs)

    def resume(self, session_name=None, session_instance=None, session_list=None, **kwargs):
        """\
        Resume a running/suspended X2Go session. 

        The L{X2GoControlSessionSTDOUT.resume()} method accepts any parameter
        that can be passed to any of the C{X2GoTerminalSession*} backend class constructors.

        @return: True if the session could be successfully resumed
        @rtype: C{bool}

        @raise X2GoUserException: if the remote user is not allowed to launch/resume X2Go sessions.

        """
        if self.get_transport() is not None:

            if not self.is_x2gouser(self.get_transport().get_username()):
                raise x2go_exceptions.X2GoUserException('remote user %s is not allowed to run X2Go commands' % self.get_transport().get_username())

            session_info = None
            try:
                if session_name is not None:
                    if session_list:
                        session_info = session_list[session_name]
                    else:
                        session_info = self.list_sessions()[session_name]
            except KeyError:
                _success = False

            _terminal = self._terminal_backend(self,
                                               profile_name=self.profile_name,
                                               session_info=session_info,
                                               info_backend=self._info_backend,
                                               list_backend=self._list_backend,
                                               proxy_backend=self._proxy_backend,
                                               client_rootdir=self.client_rootdir,
                                               session_instance=session_instance,
                                               sessions_rootdir=self.sessions_rootdir,
                                               **kwargs)

            _success = False
            try:
                if session_name is not None:
                    _success = _terminal.resume()
                else:
                    _success = _terminal.start()
            except x2go_exceptions.X2GoTerminalSessionException:
                _success = False

            if _success:
                while not _terminal.ok():
                    gevent.sleep(.2)

                if _terminal.ok():
                    self.associated_terminals[_terminal.get_session_name()] = _terminal
                    self.get_transport().reverse_tunnels[_terminal.get_session_name()] = {
                        'sshfs': (0, None),
                        'snd': (0, None),
                    }

                    return _terminal or None

        return None

    def share_desktop(self, desktop=None, user=None, display=None, share_mode=0, **kwargs):
        """\
        Share another already running desktop session. Desktop sharing can be run
        in two different modes: view-only and full-access mode.

        @param desktop: desktop ID of a sharable desktop in format C{<user>@<display>}
        @type desktop: C{str}
        @param user: user name and display number can be given separately, here give the
            name of the user who wants to share a session with you
        @type user: C{str}
        @param display: user name and display number can be given separately, here give the
            number of the display that a user allows you to be shared with
        @type display: C{str}
        @param share_mode: desktop sharing mode, 0 stands for VIEW-ONLY, 1 for  FULL-ACCESS mode
        @type share_mode: C{int}

        @return: True if the session could be successfully shared
        @rtype: C{bool}

        @raise X2GoDesktopSharingException: if C{username} and C{dislpay} do not relate to a
            sharable desktop session

        """
        if desktop:
            user = desktop.split('@')[0]
            display = desktop.split('@')[1]
        if not (user and display):
            raise x2go_exceptions.X2GoDesktopSharingException('Need user name and display number of shared desktop.')

        cmd = '%sXSHAD%sXSHAD%s' % (share_mode, user, display)

        kwargs['cmd'] = cmd
        kwargs['session_type'] = 'shared'

        return self.start(**kwargs)

    def list_desktops(self, raw=False, maxwait=20):
        """\
        List all desktop-like sessions of current user (or of users that have 
        granted desktop sharing) on the connected server.

        @param raw: if C{True}, the raw output of the server-side X2Go command 
            C{x2golistdesktops} is returned.
        @type raw: C{bool}

        @return: a list of X2Go desktops available for sharing
        @rtype: C{list}

        @raise X2GoTimeOutException: on command execution timeouts, with the server-side C{x2golistdesktops}
            command this can sometimes happen. Make sure you ignore these time-outs and to try again

        """
        if raw:
            (stdin, stdout, stderr) = self._x2go_exec_command("export HOSTNAME && x2golistdesktops")
            return stdout.read(), stderr.read()

        else:

            # this _success loop will catch errors in case the x2golistsessions output is corrupt
            # this should not be needed and is a workaround for the current X2Go server implementation

            if self.low_latency:
                maxwait = maxwait * 2

            timeout = gevent.Timeout(maxwait)
            timeout.start()
            try:
                (stdin, stdout, stderr) = self._x2go_exec_command("export HOSTNAME && x2golistdesktops")
                _stdout_read = stdout.read()
                _listdesktops = _stdout_read.split('\n')
            except gevent.timeout.Timeout:
                # if we do not get a reply here after <maxwait> seconds we will raise a time out, we have to
                # make sure that we catch this at places where we want to ignore timeouts (e.g. in the 
                # desktop list cache)
                raise x2go_exceptions.X2GoTimeOutException('x2golistdesktop command timed out')
            finally:
                timeout.cancel()

            return _listdesktops

    def list_mounts(self, session_name, raw=False, maxwait=20):
        """\
        List all mounts for a given session of the current user on the connected server.

        @param session_name: name of a session to query a list of mounts for
        @type session_name: C{str}
        @param raw: if C{True}, the raw output of the server-side X2Go command 
            C{x2golistmounts} is returned.
        @type raw: C{bool}
        @param maxwait: stop processing C{x2golistmounts} after C{<maxwait>} seconds
        @type maxwait: C{int}

        @return: a list of client-side mounts for X2Go session C{<session_name>} on the server
        @rtype: C{list}

        @raise X2GoTimeOutException: on command execution timeouts, queries with the server-side
            C{x2golistmounts} query should normally be processed quickly, a time-out may hint that the
            control session has lost its connection to the X2Go server

        """
        if raw:
            (stdin, stdout, stderr) = self._x2go_exec_command("export HOSTNAME && x2golistmounts %s" % session_name)
            return stdout.read(), stderr.read()

        else:

            if self.low_latency:
                maxwait = maxwait * 2

            # this _success loop will catch errors in case the x2golistmounts output is corrupt

            timeout = gevent.Timeout(maxwait)
            timeout.start()
            try:
                (stdin, stdout, stderr) = self._x2go_exec_command("export HOSTNAME && x2golistmounts %s" % session_name)
                _stdout_read = stdout.read()
                _listmounts = {session_name: [ line for line in _stdout_read.split('\n') if line ] }
            except gevent.timeout.Timeout:
                # if we do not get a reply here after <maxwait> seconds we will raise a time out, we have to
                # make sure that we catch this at places where we want to ignore timeouts
                raise x2go_exceptions.X2GoTimeOutException('x2golistmounts command timed out')
            finally:
                timeout.cancel()

            return _listmounts

    def list_sessions(self, raw=False):
        """\
        List all sessions of current user on the connected server.

        @param raw: if C{True}, the raw output of the server-side X2Go command 
            C{x2golistsessions} is returned.
        @type raw: C{bool}

        @return: normally an instance of a C{X2GoServerSessionList*} backend is returned. However,
            if the raw argument is set, the plain text output of the server-side C{x2golistsessions}
            command is returned
        @rtype: C{X2GoServerSessionList} instance or str

        @raise X2GoControlSessionException: on command execution timeouts, if this happens the control session will
            be interpreted as disconnected due to connection loss
        """
        if raw:
            if 'X2GO_LIST_SHADOWSESSIONS' in self._x2go_server_features:
                (stdin, stdout, stderr) = self._x2go_exec_command("export HOSTNAME && { x2golistsessions; x2golistshadowsessions; }")
            else:
                (stdin, stdout, stderr) = self._x2go_exec_command("export HOSTNAME && x2golistsessions")
            return stdout.read(), stderr.read()

        else:

            # this _success loop will catch errors in case the x2golistsessions output is corrupt
            # this should not be needed and is a workaround for the current X2Go server implementation
            _listsessions = {}
            _success = False
            _count = 0
            _maxwait = 20

            # we will try this 20 times before giving up... we might simply catch the x2golistsessions
            # output in the middle of creating a session in the database...
            while not _success and _count < _maxwait:
                _count += 1
                try:
                    if 'X2GO_LIST_SHADOWSESSIONS' in self._x2go_server_features:
                        (stdin, stdout, stderr) = self._x2go_exec_command("export HOSTNAME && { x2golistsessions; x2golistshadowsessions; }")
                    else:
                        (stdin, stdout, stderr) = self._x2go_exec_command("export HOSTNAME && x2golistsessions")
                    _stdout_read = stdout.read()
                    _listsessions = self._list_backend(_stdout_read, info_backend=self._info_backend).sessions
                    _success = True
                except KeyError:
                    gevent.sleep(1)
                except IndexError:
                    gevent.sleep(1)
                except ValueError:
                    gevent.sleep(1)

            if _count >= _maxwait:
                self.session_died = True
                self.disconnect()
                raise x2go_exceptions.X2GoControlSessionException('x2golistsessions command failed after we have tried 20 times')

            # update internal variables when list_sessions() is called
            for _session_name, _terminal in self.associated_terminals.items():
                if _session_name in _listsessions.keys():
                    # update the whole session_info object within the terminal session
                    if hasattr(self.associated_terminals[_session_name], 'session_info') and not self.associated_terminals[_session_name].is_session_info_protected():
                        self.associated_terminals[_session_name].session_info.update(_listsessions[_session_name])
                else:
                    try: del self.associated_terminals[_session_name]
                    except KeyError: pass
                    self.terminated_terminals.append(_session_name)
                if _terminal.is_suspended():
                    try: del self.associated_terminals[_session_name]
                    except KeyError: pass


            return _listsessions

    def clean_sessions(self, destroy_terminals=True, published_applications=False):
        """\
        Find X2Go terminals that have previously been started by the
        connected user on the remote X2Go server and terminate them.

        @param destroy_terminals: destroy the terminal session instances after cleanup
        @type destroy_terminals: C{bool}
        @param published_applications: also clean up published applications providing sessions
        @type published_applications: C{bool}

        """
        session_list = self.list_sessions()
        if published_applications:
            session_names = session_list.keys()
        else:
            session_names = [ _sn for _sn in session_list.keys() if not session_list[_sn].is_published_applications_provider() ]
        for session_name in session_names:
            self.terminate(session_name=session_name, destroy_terminals=destroy_terminals)

    def is_connected(self):
        """\
        Returns C{True} if this control session is connected to the remote server (that
        is: if it has a valid Paramiko/SSH transport object).

        @return: X2Go session connected?
        @rtype: C{bool}

        """
        return self.get_transport() is not None and self.get_transport().is_authenticated()

    def is_running(self, session_name):
        """\
        Returns C{True} if the given X2Go session is in running state,
        C{False} else.

        @param session_name: X2Go name of the session to be queried
        @type session_name: C{str}

        @return: X2Go session running? If C{<session_name>} is not listable by the L{list_sessions()} method then C{None} is returned
        @rtype: C{bool} or C{None}

        """
        session_infos = self.list_sessions()
        if session_name in session_infos.keys():
            return session_infos[session_name].is_running()
        return None

    def is_suspended(self, session_name):
        """\
        Returns C{True} if the given X2Go session is in suspended state,
        C{False} else.

        @return: X2Go session suspended? If C{<session_name>} is not listable by the L{list_sessions()} method then C{None} is returned
        @rtype: C{bool} or C{None}

        """
        session_infos = self.list_sessions()
        if session_name in session_infos.keys():
            return session_infos[session_name].is_suspended()
        return None

    def has_terminated(self, session_name):
        """\
        Returns C{True} if the X2Go session with name C{<session_name>} has been seen
        by this control session and--in the meantime--has been terminated.

        If C{<session_name>} has not been seen, yet, the method will return C{None}.

        @return: X2Go session has terminated?
        @rtype: C{bool} or C{None}

        """
        session_infos = self.list_sessions()
        if session_name in self.terminated_terminals:
            return True
        if session_name not in session_infos.keys() and session_name in self.associated_terminals.keys():
            # do a post-mortem tidy up
            self.terminate(session_name)
            return True
        if self.is_suspended(session_name) or self.is_running(session_name):
            return False

        return None

    def suspend(self, session_name):
        """\
        Suspend X2Go session with name C{<session_name>} on the connected
        server.

        @param session_name: X2Go name of the session to be suspended
        @type session_name: C{str}

        @return: C{True} if the session could be successfully suspended
        @rtype: C{bool}

        """
        _ret = False
        _session_names = [ t.get_session_name() for t in self.associated_terminals.values() ]
        if session_name in _session_names:

            self.logger('suspending associated terminal session: %s' % session_name, loglevel=log.loglevel_DEBUG)
            (stdin, stdout, stderr) = self._x2go_exec_command("x2gosuspend-session %s" % session_name, loglevel=log.loglevel_DEBUG)
            stdout.read()
            stderr.read()
            if self.associated_terminals.has_key(session_name):
                if self.associated_terminals[session_name] is not None:
                    self.associated_terminals[session_name].__del__()
                try: del self.associated_terminals[session_name]
                except KeyError: pass
            _ret = True

        else:

            self.logger('suspending non-associated terminal session: %s' % session_name, loglevel=log.loglevel_DEBUG)
            (stdin, stdout, stderr) = self._x2go_exec_command("x2gosuspend-session %s" % session_name, loglevel=log.loglevel_DEBUG)
            stdout.read()
            stderr.read()
            _ret = True

        return _ret

    def terminate(self, session_name, destroy_terminals=True):
        """\
        Terminate X2Go session with name C{<session_name>} on the connected
        server.

        @param session_name: X2Go name of the session to be terminated
        @type session_name: C{str}

        @return: C{True} if the session could be successfully terminated
        @rtype: C{bool}

        """

        _ret = False
        _session_names = [ t.get_session_name() for t in self.associated_terminals.values() ]
        if session_name in _session_names:

            self.logger('terminating associated session: %s' % session_name, loglevel=log.loglevel_DEBUG)
            (stdin, stdout, stderr) = self._x2go_exec_command("x2goterminate-session %s" % session_name, loglevel=log.loglevel_DEBUG)
            stdout.read()
            stderr.read()
            if self.associated_terminals.has_key(session_name):
                if self.associated_terminals[session_name] is not None and destroy_terminals:
                    self.associated_terminals[session_name].__del__()
                try: del self.associated_terminals[session_name]
                except KeyError: pass
            self.terminated_terminals.append(session_name)
            _ret = True

        else:

            self.logger('terminating non-associated session: %s' % session_name, loglevel=log.loglevel_DEBUG)
            (stdin, stdout, stderr) = self._x2go_exec_command("x2goterminate-session %s" % session_name, loglevel=log.loglevel_DEBUG)
            stdout.read()
            stderr.read()
            _ret = True

        return _ret



# -*- coding: utf-8 -*-

# Copyright (C) 2010-2013 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
X2GoSessionRegistry class - the X2GoClient's session registry backend

"""
__NAME__ = 'x2gosessregistry-pylib'

import os
import copy
import types
import time
import threading
import re

# Python X2Go modules
import log
import utils
import session
import x2go_exceptions

# import the default terminal session backend
from x2go.backends.control import X2GoControlSession as _X2GoControlSession
from x2go.backends.terminal import X2GoTerminalSession as _X2GoTerminalSession
from x2go.backends.info import X2GoServerSessionInfo as _X2GoServerSessionInfo
from x2go.backends.info import X2GoServerSessionList as _X2GoServerSessionList
from x2go.backends.proxy import X2GoProxy as _X2GoProxy
from x2go.backends.settings import X2GoClientSettings as _X2GoClientSettings
from x2go.backends.printing import X2GoClientPrinting as _X2GoClientPrinting

from defaults import LOCAL_HOME as _LOCAL_HOME
from defaults import X2GO_CLIENT_ROOTDIR as _X2GO_CLIENT_ROOTDIR
from defaults import X2GO_SESSIONS_ROOTDIR as _X2GO_SESSIONS_ROOTDIR
from defaults import X2GO_SESSIONPROFILE_DEFAULTS as _X2GO_SESSIONPROFILE_DEFAULTS
from defaults import X2GO_SSH_ROOTDIR as _X2GO_SSH_ROOTDIR

class X2GoSessionRegistry(object):
    """\
    This class is utilized by L{X2GoClient} instances to maintain a good overview on
    session status of all associated L{X2GoSession} instances.

    """
    def __init__(self, client_instance,
                 logger=None, loglevel=log.loglevel_DEFAULT):
        """\
        @param client_instance: the L{X2GoClient} instance that instantiated this L{X2GoSessionRegistry} instance.
        @type client_instance: L{X2GoClient} instance
        @param logger: you can pass an L{X2GoLogger} object to the L{X2GoClientXConfig} constructor
        @type logger: C{obj}
        @param loglevel: if no L{X2GoLogger} object has been supplied a new one will be
            constructed with the given loglevel
        @type loglevel: C{int}

        """
        if logger is None:
            self.logger = log.X2GoLogger(loglevel=loglevel)
        else:
            self.logger = copy.deepcopy(logger)
        self.logger.tag = __NAME__

        self.client_instance = client_instance

        self.registry = {}
        self.control_sessions = {}
        self.master_sessions = {}

        self._last_available_session_registration = None
        self._skip_auto_registration = False
        self._profile_locks = {}

    def keys(self):
        """\
        A list of session registry keys.

        @return: session registry key list
        @rtype: C{list}

        """
        return self.registry.keys()

    def __repr__(self):
        result = 'X2GoSessionRegistry('
        for p in dir(self):
            if '__' in p or not p in self.__dict__ or type(p) is types.InstanceType: continue
            result += p + '=' + str(self.__dict__[p]) + ','
        result = result.strip(',')
        return result + ')'

    def __call__(self, session_uuid):
        """\
        Returns the L{X2GoSession} instance for a given session UUID hash.

        @param session_uuid: the X2Go session's UUID registry hash
        @type session_uuid: C{str}

        @return: the corresponding L{X2GoSession} instance
        @rtype: L{X2GoSession} instance

        @raise X2GoSessionRegistryException: if the given session UUID could not be found

        """
        try:
            return self.registry[session_uuid]
        except KeyError:
            raise x2go_exceptions.X2GoSessionRegistryException('No session found for UUID %s' % session_uuid)

    def disable_session_auto_registration(self):
        """\
        This method is used to temporarily skip auto-registration of newly appearing
        X2Go session on the server side. This is necessary during session startups to
        assure that the session registry does not get filled with session UUID 
        duplicates.

        """
        self._skip_auto_registration = True

    def enable_session_auto_registration(self):
        """\
        This method is used to temporarily (re-)enable auto-registration of newly appearing
        X2Go session on the server side.

        """
        self._skip_auto_registration = False

    def forget(self, session_uuid):
        """\
        Forget the complete record for session UUID C{session_uuid}.

        @param session_uuid: the X2Go session's UUID registry hash
        @type session_uuid: C{str}

        """
        try:
            del self.registry[session_uuid]
            self.logger('Forgetting session UUID %s' % session_uuid, loglevel=log.loglevel_DEBUG)
        except KeyError:
            pass

    def get_profile_id(self, session_uuid):
        """\
        Retrieve the profile ID of a given session UUID hash.

        @param session_uuid: the X2Go session's UUID registry hash
        @type session_uuid: C{str}

        @return: profile ID
        @rtype: C{str}

        """
        return self(session_uuid).get_profile_id()

    def get_profile_name(self, session_uuid):
        """\
        Retrieve the profile name of a given session UUID hash.

        @param session_uuid: the X2Go session's UUID registry hash
        @type session_uuid: C{str}

        @return: profile name
        @rtype: C{str}

        """
        return self(session_uuid).get_profile_name()

    def session_summary(self, session_uuid, status_only=False):
        """\
        Compose a session summary (as Python dictionary).

        @param session_uuid: the X2Go session's UUID registry hash
        @type session_uuid: C{str}

        @return: session summary dictionary
        @rtype: C{dict}

        """
        _session_summary = {}
        _r = False
        if session_uuid in [ s() for s in self.registered_sessions() ]:
            _r = True

        if not status_only:
            _session_summary['uuid'] = _r and session_uuid or None
            _session_summary['profile_id'] = _r and self.get_profile_id(session_uuid) or ''
            _session_summary['profile_name'] = _r and self.get_profile_name(session_uuid) or ''
            _session_summary['session_name'] = _r and self(session_uuid).get_session_name() or ''
            _session_summary['control_session'] = _r and self(session_uuid).get_control_session() or None
            _session_summary['control_params'] = _r and self(session_uuid).control_params or {}
            _session_summary['terminal_session'] = _r and self(session_uuid).get_terminal_session() or None
            _session_summary['terminal_params'] = _r and self(session_uuid).terminal_params or {}
            _session_summary['active_threads'] = _r and bool(self(session_uuid).get_terminal_session()) and self(session_uuid).get_terminal_session().active_threads or []
            _session_summary['backends'] = {
                'control': _r and self(session_uuid).control_backend or None,
                'terminal': _r and self(session_uuid).terminal_backend or None,
                'info': _r and self(session_uuid).info_backend or None,
                'list': _r and self(session_uuid).list_backend or None,
                'proxy': _r and self(session_uuid).proxy_backend or None,
            }

        if _r:
            _session_summary['virgin'] = self(session_uuid).virgin
            _session_summary['connected'] = self(session_uuid).connected
            _session_summary['running'] = self(session_uuid).running
            _session_summary['suspended'] = self(session_uuid).suspended
            _session_summary['terminated'] = self(session_uuid).terminated
        else:
            _session_summary['virgin'] = None
            _session_summary['connected'] = None
            _session_summary['running'] = None
            _session_summary['suspended'] = None
            _session_summary['terminated'] = None
        return _session_summary

    def update_status(self, session_uuid=None, profile_name=None, profile_id=None, session_list=None, force_update=False, newly_connected=False):
        """\
        Update the session status for L{X2GoSession} that is represented by a given session UUID hash,
        profile name or profile ID.

        @param session_uuid: the X2Go session's UUID registry hash
        @type session_uuid: C{str}
        @param profile_name: alternatively, a profile name can be specified (the stati of all registered sessions for this session
            profile will be updated)
        @type profile_name: C{str}
        @param profile_id: alternatively, a profile ID can be given (the stati of all registered sessions for this session 
            profile will be updated)
        @type profile_id: C{str}
        @param session_list: an optional C{X2GoServerSessionList*} instance (as returned by the L{X2GoClient.list_sessions()} command can
            be passed to this method.
        @type session_list: C{X2GoServerSessionList*} instance
        @param force_update: make sure the session status gets really updated
        @type force_update: C{bool}

        @return: C{True} if this method has been successful
        @rtype: C{bool}

        @raise X2GoSessionRegistryException: if the combination of C{session_uuid}, C{profile_name} and C{profile_id} does not match the requirement: 
            only one of them

        """
        if session_uuid and profile_name or session_uuid and profile_id or profile_name and profile_id:
            raise x2go_exceptions.X2GoSessionRegistryException('only one of the possible method parameters is allowed (session_uuid, profile_name or profile_id)')
        elif session_uuid is None and profile_name is None and profile_id is None:
            raise x2go_exceptions.X2GoSessionRegistryException('at least one of the method parameters session_uuid, profile_name or profile_id must be given')

        if session_uuid:
            session_uuids = [ session_uuid ]
        elif profile_name:
            session_uuids = [ s() for s in self.registered_sessions_of_profile_name(profile_name, return_objects=True) ]
        elif profile_id:
            session_uuids = [ s() for s in self.registered_sessions_of_profile_name(self.client_instance.to_profile_name(profile_id), return_objects=True) ]

        for _session_uuid in session_uuids:

            # only operate on instantiated X2GoSession objects
            if type(self(_session_uuid)) != session.X2GoSession:
                continue

            if self(_session_uuid).is_locked():
                continue

            if not self(_session_uuid).update_status(session_list=session_list, force_update=force_update):
                # skip this run, as nothing has changed since the last time...
                continue

            _last_status = copy.deepcopy(self(_session_uuid)._last_status)
            _current_status = copy.deepcopy(self(_session_uuid)._current_status)

            # at this point we hook into the X2GoClient instance and call notification methods
            # that can be used to inform an application that something has happened

            _profile_name = self(_session_uuid).get_profile_name()
            _session_name = self(_session_uuid).get_session_name()

            if self(_session_uuid).get_server_hostname() != _current_status['server']:

                # if the server (hostname) has changed due to a configuration change we skip all notifications
                self(_session_uuid).session_cleanup()
                self(_session_uuid).__del__()
                if len(self.virgin_sessions_of_profile_name(profile_name)) > 1:
                    del self.registry[_session_uuid]

            elif not _last_status['running'] and _current_status['running'] and not _current_status['faulty']:
                # session has started
                if newly_connected:
                    # from a suspended state
                    self.client_instance.HOOK_on_found_session_running_after_connect(session_uuid=_session_uuid, profile_name=_profile_name, session_name=_session_name)
                else:
                    # explicitly ask for the terminal_session object directly here, so we also get 'PENDING' terminal sessions here...
                    if self(_session_uuid).terminal_session:

                        # declare as master session if appropriate
                        if _profile_name not in self.master_sessions.keys():
                            self.master_sessions[_profile_name] = self(_session_uuid)
                            self(_session_uuid).set_master_session()

                        elif (not self.master_sessions[_profile_name].is_desktop_session() and self(_session_uuid).is_desktop_session()) or \
                             (not self.master_sessions[_profile_name].is_desktop_session() and self(_session_uuid).is_published_applications_provider()):
                                self(self.master_sessions[_profile_name]()).unset_master_session()
                                self.master_sessions[_profile_name] = self(_session_uuid)
                                self(_session_uuid).set_master_session()

                        if _last_status['suspended']:
                            # from a suspended state
                            self.client_instance.HOOK_on_session_has_resumed_by_me(session_uuid=_session_uuid, profile_name=_profile_name, session_name=_session_name)
                        elif _last_status['virgin']:
                            # as a new session
                            self.client_instance.HOOK_on_session_has_started_by_me(session_uuid=_session_uuid, profile_name=_profile_name, session_name=_session_name)

                    else:
                        if _last_status['suspended']:
                            # from a suspended state
                            self.client_instance.HOOK_on_session_has_resumed_by_other(session_uuid=_session_uuid, profile_name=_profile_name, session_name=_session_name)
                        elif _last_status['connected'] and _last_status['virgin']:
                            # as a new session, do not report directly after connect due to many false positives then...
                            self.client_instance.HOOK_on_session_has_started_by_other(session_uuid=_session_uuid, profile_name=_profile_name, session_name=_session_name)

            elif _last_status['connected'] and (not _last_status['suspended'] and _current_status['suspended']) and not _current_status['faulty'] and _session_name:

                # unregister as master session
                if _profile_name in self.master_sessions.keys():
                    if self.master_sessions[_profile_name] == self(_session_uuid):

                        self(_session_uuid).unset_master_session()
                        del self.master_sessions[_profile_name]

                # session has been suspended
                self(_session_uuid).session_cleanup()
                self.client_instance.HOOK_on_session_has_been_suspended(session_uuid=_session_uuid, profile_name=_profile_name, session_name=_session_name)

            elif _last_status['connected'] and (not _last_status['terminated'] and _current_status['terminated']) and not _current_status['faulty'] and _session_name:

                # unregister as master session
                if _profile_name in self.master_sessions.keys():
                    if self.master_sessions[_profile_name] == self(_session_uuid):

                        self(_session_uuid).unset_master_session()
                        del self.master_sessions[_profile_name]

                # session has terminated
                self.client_instance.HOOK_on_session_has_terminated(session_uuid=_session_uuid, profile_name=_profile_name, session_name=_session_name)
                try: self(_session_uuid).session_cleanup()
                except x2go_exceptions.X2GoSessionException: pass
                try: self(_session_uuid).__del__()
                except x2go_exceptions.X2GoSessionException: pass
                if len(self.virgin_sessions_of_profile_name(profile_name)) > 1:
                    self.forget(_session_uuid)

        # detect master sessions for connected profiles that have lost (suspend/terminate) their master session or never had a master session
        for _profile_name in [ p for p in self.connected_profiles(return_profile_names=True) if p not in self.master_sessions.keys() ]:
            _running_associated_sessions = [ _s for _s in self.running_sessions_of_profile_name(_profile_name, return_objects=True) if _s.is_associated() ]
            if _running_associated_sessions:
                for _r_a_s in _running_associated_sessions:
                    if _r_a_s.is_desktop_session():
                        self.master_sessions[_profile_name] = _r_a_s
                        _r_a_s.set_master_session(wait=1)
                        break
                if not self.master_sessions.has_key(_profile_name):
                    _pubapp_associated_sessions = self.pubapp_sessions_of_profile_name(_profile_name, return_objects=True)
                    if _pubapp_associated_sessions:
                        self.master_sessions[_profile_name] = _pubapp_associated_sessions[0]
                        _pubapp_associated_sessions[0].set_master_session(wait=2)
                    else:
                        self.master_sessions[_profile_name] = _running_associated_sessions[0]
                        _running_associated_sessions[0].set_master_session(wait=2)

        return True

    def register_available_server_sessions(self, profile_name, session_list=None, newly_connected=False, re_register=False, skip_pubapp_sessions=False):
        """\
        Register server-side available X2Go sessions with this L{X2GoSessionRegistry} instance for a given profile name.

        @param profile_name: session profile name to register available X2Go sessions for
        @type profile_name: C{str}
        @param session_list: an optional C{X2GoServerSessionList*} instance (as returned by the L{X2GoClient.list_sessions()} command can
            be passed to this method.
        @type session_list: C{X2GoServerSessionList*} instance
        @param newly_connected: give a hint that the session profile got newly connected
        @type newly_connected: C{bool}
        @param re_register: re-register available sessions, needs to be done after changes to the session profile
        @type re_register: C{bool}
        @param skip_pubapp_sessions: Do not register published applications sessions
        @type skip_pubapp_sessions: C{bool}

        """
        if self._last_available_session_registration is not None:
            _now = time.time()
            _time_delta = _now - self._last_available_session_registration
            if _time_delta < 2 and not re_register:
                self.logger('registration interval too short (%s), skipping automatic session registration...' % _time_delta, loglevel=log.loglevel_DEBUG)
                return
            self._last_available_session_registration = _now

        _connected_sessions = self.connected_sessions_of_profile_name(profile_name=profile_name, return_objects=False)
        _registered_sessions = self.registered_sessions_of_profile_name(profile_name=profile_name, return_objects=False)
        _session_names = [ self(s_uuid).session_name for s_uuid in _registered_sessions if self(s_uuid).session_name is not None ]

        if _connected_sessions:
            # any of the connected sessions is valuable for accessing the profile's control 
            # session commands, so we simply take the first that comes in...
            _ctrl_session = self(_connected_sessions[0])

            if session_list is None:
                session_list = _ctrl_session.list_sessions()

            # make sure the session registry gets updated before registering new session
            # (if the server name has changed, this will kick out obsolete X2GoSessions)
            self.update_status(profile_name=profile_name, session_list=session_list, force_update=True)
            for session_name in session_list.keys():
                if (session_name not in _session_names and not self._skip_auto_registration) or re_register:
                    server = _ctrl_session.get_server_hostname()
                    profile_id = _ctrl_session.get_profile_id()

                    # reconstruct all session options of _ctrl_session to auto-register a suspended session
                    # found on the _ctrl_session's connected server
                    _clone_kwargs = _ctrl_session.__dict__
                    kwargs = {}
                    kwargs.update(self.client_instance.session_profiles.to_session_params(profile_id))
                    kwargs['client_instance'] = self.client_instance
                    kwargs['control_backend'] = _clone_kwargs['control_backend']
                    kwargs['terminal_backend'] = _clone_kwargs['terminal_backend']
                    kwargs['proxy_backend'] = _clone_kwargs['proxy_backend']
                    kwargs['info_backend'] = _clone_kwargs['info_backend']
                    kwargs['list_backend'] = _clone_kwargs['list_backend']
                    kwargs['settings_backend'] = _clone_kwargs['settings_backend']
                    kwargs['printing_backend'] = _clone_kwargs['printing_backend']
                    kwargs['keep_controlsession_alive'] = _clone_kwargs['keep_controlsession_alive']
                    kwargs['client_rootdir'] = _clone_kwargs['client_rootdir']
                    kwargs['sessions_rootdir'] = _clone_kwargs['sessions_rootdir']

                    try: del kwargs['server'] 
                    except: pass
                    try: del kwargs['profile_name']
                    except: pass
                    try: del kwargs['profile_id'] 
                    except: pass

                    # this if clause catches problems when x2golistsessions commands give weird results
                    if not self.has_session_of_session_name(session_name) or re_register:
                        if not (skip_pubapp_sessions and re.match('.*_stRPUBLISHED_.*', session_name)):
                            session_uuid = self.register(server, profile_id, profile_name,
                                                         session_name=session_name, virgin=False,
                                                         **kwargs
                                                        )
                            self(session_uuid).connected = True
                            self.update_status(session_uuid=session_uuid, force_update=True, newly_connected=newly_connected)

    def register(self, server, profile_id, profile_name,
                 session_name=None,
                 control_backend=_X2GoControlSession,
                 terminal_backend=_X2GoTerminalSession,
                 info_backend=_X2GoServerSessionInfo,
                 list_backend=_X2GoServerSessionList,
                 proxy_backend=_X2GoProxy,
                 settings_backend=_X2GoClientSettings,
                 printing_backend=_X2GoClientPrinting,
                 client_rootdir=os.path.join(_LOCAL_HOME,_X2GO_CLIENT_ROOTDIR),
                 sessions_rootdir=os.path.join(_LOCAL_HOME,_X2GO_SESSIONS_ROOTDIR),
                 ssh_rootdir=os.path.join(_LOCAL_HOME,_X2GO_SSH_ROOTDIR),
                 keep_controlsession_alive=True,
                 add_to_known_hosts=False,
                 known_hosts=None,
                 **kwargs):
        """\
        Register a new L{X2GoSession} instance with this L{X2GoSessionRegistry}.

        @param server: hostname of X2Go server
        @type server: C{str}
        @param profile_id: profile ID
        @type profile_id: C{str}
        @param profile_name: profile name
        @type profile_name: C{str}
        @param session_name: session name (if available)
        @type session_name: C{str}
        @param control_backend: X2Go control session backend to use
        @type control_backend: C{class}
        @param terminal_backend: X2Go terminal session backend to use
        @type terminal_backend: C{class}
        @param info_backend: X2Go session info backend to use
        @type info_backend: C{class}
        @param list_backend: X2Go session list backend to use
        @type list_backend: C{class}
        @param proxy_backend: X2Go proxy backend to use
        @type proxy_backend: C{class}
        @param settings_backend: X2Go client settings backend to use
        @type settings_backend: C{class}
        @param printing_backend: X2Go client printing backend to use
        @type printing_backend: C{class}
        @param client_rootdir: client base dir (default: ~/.x2goclient)
        @type client_rootdir: C{str}
        @param sessions_rootdir: sessions base dir (default: ~/.x2go)
        @type sessions_rootdir: C{str}
        @param ssh_rootdir: ssh base dir (default: ~/.ssh)
        @type ssh_rootdir: C{str}
        @param keep_controlsession_alive: On last L{X2GoSession.disconnect()} keep the associated C{X2GoControlSession*} instance alive?
        @ŧype keep_controlsession_alive: C{bool}
        @param add_to_known_hosts: Auto-accept server host validity?
        @type add_to_known_hosts: C{bool}
        @param known_hosts: the underlying Paramiko/SSH systems C{known_hosts} file
        @type known_hosts: C{str}
        @param kwargs: all other options will be passed on to the constructor of the to-be-instantiated L{X2GoSession} instance
        @type C{dict}

        @return: the session UUID of the newly registered (or re-registered) session
        @rtype: C{str}

        """
        if profile_id not in self._profile_locks.keys():
            self._profile_locks[profile_id] = threading.Lock()

        self._profile_locks[profile_id].acquire()

        control_session = None
        if profile_id in self.control_sessions.keys():
            control_session = self.control_sessions[profile_id]

        try:
            _params = self.client_instance.session_profiles.to_session_params(profile_id)

        except x2go_exceptions.X2GoProfileException:
            _params = utils._convert_SessionProfileOptions_2_SessionParams(_X2GO_SESSIONPROFILE_DEFAULTS)

        for _k in _params.keys():
            if _k in kwargs.keys():
                _params[_k] = kwargs[_k]

        # when starting a new session, we will try to use unused registered virgin sessions
        # depending on your application layout, there should either be one or no such virgin session at all
        _virgin_sessions = [ s for s in self.virgin_sessions_of_profile_name(profile_name, return_objects=True) if not s.activated ]
        if _virgin_sessions and not session_name:
            session_uuid = _virgin_sessions[0].get_uuid()
            self(session_uuid).activated = True
            self.logger('using already initially-registered yet-unused session %s' % session_uuid, loglevel=log.loglevel_NOTICE)

        else:
            session_uuid = self.get_session_of_session_name(session_name, match_profile_name=profile_name)
            if session_uuid is not None: self.logger('using already registered-by-session-name session %s' % session_uuid, loglevel=log.loglevel_NOTICE)

        if session_uuid is not None:
            self(session_uuid).activated = True
            self(session_uuid).update_params(_params)
            self(session_uuid).set_server(server)
            self(session_uuid).set_profile_name(profile_name)
            self._profile_locks[profile_id].release()
            return session_uuid

        try: del _params['server'] 
        except: pass
        try: del _params['profile_name']
        except: pass
        try: del _params['profile_id'] 
        except: pass

        s = session.X2GoSession(server=server, control_session=control_session,
                                profile_id=profile_id, profile_name=profile_name,
                                session_name=session_name,
                                control_backend=control_backend,
                                terminal_backend=terminal_backend,
                                info_backend=info_backend,
                                list_backend=list_backend,
                                proxy_backend=proxy_backend,
                                settings_backend=settings_backend,
                                printing_backend=printing_backend,
                                client_rootdir=client_rootdir,
                                sessions_rootdir=sessions_rootdir,
                                ssh_rootdir=ssh_rootdir,
                                keep_controlsession_alive=keep_controlsession_alive,
                                add_to_known_hosts=add_to_known_hosts,
                                known_hosts=known_hosts,
                                client_instance=self.client_instance,
                                logger=self.logger, **_params)

        session_uuid = s._X2GoSession__get_uuid()
        self.logger('registering X2Go session %s...' % profile_name, log.loglevel_NOTICE)
        self.logger('registering X2Go session with UUID %s' % session_uuid, log.loglevel_DEBUG)

        self.registry[session_uuid] = s
        if profile_id not in self.control_sessions.keys():
            self.control_sessions[profile_id] = s.get_control_session()

        # make sure a new session is a non-master session unless promoted in update_status method
        self(session_uuid).unset_master_session()
        if control_session is None:
            self(session_uuid).do_auto_connect()

        self._profile_locks[profile_id].release()
        return session_uuid

    def has_session_of_session_name(self, session_name, match_profile_name=None):
        """\
        Detect if we know about an L{X2GoSession} of name C{<session_name>}.

        @param session_name: name of session to be searched for
        @type session_name: C{str}
        @param match_profile_name: a session's profile_name must match this profile name
        @type match_profile_name: C{str}

        @return: C{True} if a session of C{<session_name>} has been found
        @rtype: C{bool}

        """
        return bool(self.get_session_of_session_name(session_name, match_profile_name=match_profile_name))

    def get_session_of_session_name(self, session_name, return_object=False, match_profile_name=None):
        """\
        Retrieve the L{X2GoSession} instance with session name C{<session_name>}.

        @param session_name: name of session to be retrieved
        @type session_name: C{str}
        @param return_object: if C{False} the session UUID hash will be returned, if C{True} the L{X2GoSession} instance will be returned
        @type return_object: C{bool}
        @param match_profile_name: returned sessions must match this profile name
        @type match_profile_name: C{str}

        @return: L{X2GoSession} object or its representing session UUID hash
        @rtype: L{X2GoSession} instance or C{str}

        @raise X2GoSessionRegistryException: if there is more than one L{X2GoSession} registered for C{<session_name>} within
            the same L{X2GoClient} instance. This should never happen!

        """
        if match_profile_name is None:
            reg_sessions = self.registered_sessions()
        else:
            reg_sessions = self.registered_sessions_of_profile_name(match_profile_name)
        found_sessions = [ s for s in reg_sessions if s.session_name == session_name and s.session_name is not None ]
        if len(found_sessions) == 1:
            session = found_sessions[0]
            if return_object:
                return session
            else:
                return session.get_uuid()
        elif len(found_sessions) > 1:
            raise x2go_exceptions.X2GoSessionRegistryException('there should only be one registered session of name ,,%s\'\'' % session_name)
        else:
            return None

    def _sessionsWithState(self, state, return_objects=True, return_profile_names=False, return_profile_ids=False, return_session_names=False):
        if state == 'associated':
            sessions = [ ts for ts in self.registry.values() if ts.has_terminal_session() ]
        elif state == 'registered':
            sessions = [ ts for ts in self.registry.values() ]
        else:
            sessions = [ ts for ts in self.registry.values() if eval('ts.%s' % state) ]
        if return_profile_names:
            profile_names = []
            for session in sessions:
                if session.profile_name not in profile_names:
                    profile_names.append(session.profile_name)
            return profile_names
        elif return_profile_ids:
            profile_ids = []
            for session in sessions:
                if session.profile_id not in profile_ids:
                    profile_ids.append(session.profile_id)
            return profile_ids
        elif return_session_names:
            session_names = []
            for session in sessions:
                if session.session_name not in session_names:
                    session_names.append(session.session_name)
            return session_names
        elif return_objects:
            return sessions
        else:
            return [s.get_uuid() for s in sessions ]

    def connected_sessions(self, return_objects=True, return_profile_names=False, return_profile_ids=False, return_session_names=False):
        """\
        Retrieve a list of sessions that the underlying L{X2GoClient} instances is currently connected to.
        If none of the C{return_*} options is specified a list of session UUID hashes will be returned.

        @param return_objects: return as list of L{X2GoSession} instances
        @type return_objects: C{bool}
        @param return_profile_names: return as list of profile names
        @type return_profile_names: C{bool}
        @param return_profile_ids: return as list of profile IDs
        @type return_profile_ids: C{bool}
        @param return_session_names: return as list of X2Go session names
        @type return_session_names: C{bool}

        @return: a session list (as UUID hashes, objects, profile names/IDs or session names)
        @rtype: C{list}

        """
        return self._sessionsWithState('connected', return_objects=return_objects, return_profile_names=return_profile_names, return_profile_ids=return_profile_ids, return_session_names=return_session_names)

    def associated_sessions(self, return_objects=True, return_profile_names=False, return_profile_ids=False, return_session_names=False):
        """\
        Retrieve a list of sessions that are currently associated by an C{X2GoTerminalSession*} to the underlying L{X2GoClient} instance.
        If none of the C{return_*} options is specified a list of session UUID hashes will be returned.

        @param return_objects: return as list of L{X2GoSession} instances
        @type return_objects: C{bool}
        @param return_profile_names: return as list of profile names
        @type return_profile_names: C{bool}
        @param return_profile_ids: return as list of profile IDs
        @type return_profile_ids: C{bool}
        @param return_session_names: return as list of X2Go session names
        @type return_session_names: C{bool}

        @return: a session list (as UUID hashes, objects, profile names/IDs or session names)
        @rtype: C{list}

        """
        return self._sessionsWithState('associated', return_objects=return_objects, return_profile_names=return_profile_names, return_profile_ids=return_profile_ids, return_session_names=return_session_names)

    def virgin_sessions(self, return_objects=True, return_profile_names=False, return_profile_ids=False, return_session_names=False):
        """\
        Retrieve a list of sessions that are currently still in virgin state (not yet connected, associated etc.).
        If none of the C{return_*} options is specified a list of session UUID hashes will be returned.

        @param return_objects: return as list of L{X2GoSession} instances
        @type return_objects: C{bool}
        @param return_profile_names: return as list of profile names
        @type return_profile_names: C{bool}
        @param return_profile_ids: return as list of profile IDs
        @type return_profile_ids: C{bool}
        @param return_session_names: return as list of X2Go session names
        @type return_session_names: C{bool}

        @return: a session list (as UUID hashes, objects, profile names/IDs or session names)
        @rtype: C{list}

        """
        return self._sessionsWithState('virgin', return_objects=return_objects, return_profile_names=return_profile_names, return_profile_ids=return_profile_ids, return_session_names=return_session_names)

    def running_sessions(self, return_objects=True, return_profile_names=False, return_profile_ids=False, return_session_names=False):
        """\
        Retrieve a list of sessions that are currently in running state.
        If none of the C{return_*} options is specified a list of session UUID hashes will be returned.

        @param return_objects: return as list of L{X2GoSession} instances
        @type return_objects: C{bool}
        @param return_profile_names: return as list of profile names
        @type return_profile_names: C{bool}
        @param return_profile_ids: return as list of profile IDs
        @type return_profile_ids: C{bool}
        @param return_session_names: return as list of X2Go session names
        @type return_session_names: C{bool}

        @return: a session list (as UUID hashes, objects, profile names/IDs or session names)
        @rtype: C{list}

        """
        return self._sessionsWithState('running', return_objects=return_objects, return_profile_names=return_profile_names, return_profile_ids=return_profile_ids, return_session_names=return_session_names)

    def suspended_sessions(self, return_objects=True, return_profile_names=False, return_profile_ids=False, return_session_names=False):
        """\
        Retrieve a list of sessions that are currently in suspended state.
        If none of the C{return_*} options is specified a list of session UUID hashes will be returned.

        @param return_objects: return as list of L{X2GoSession} instances
        @type return_objects: C{bool}
        @param return_profile_names: return as list of profile names
        @type return_profile_names: C{bool}
        @param return_profile_ids: return as list of profile IDs
        @type return_profile_ids: C{bool}
        @param return_session_names: return as list of X2Go session names
        @type return_session_names: C{bool}

        @return: a session list (as UUID hashes, objects, profile names/IDs or session names)
        @rtype: C{list}

        """
        return self._sessionsWithState('suspended', return_objects=return_objects, return_profile_names=return_profile_names, return_profile_ids=return_profile_ids, return_session_names=return_session_names)

    def terminated_sessions(self, return_objects=True, return_profile_names=False, return_profile_ids=False, return_session_names=False):
        """\
        Retrieve a list of sessions that have terminated recently.
        If none of the C{return_*} options is specified a list of session UUID hashes will be returned.

        @param return_objects: return as list of L{X2GoSession} instances
        @type return_objects: C{bool}
        @param return_profile_names: return as list of profile names
        @type return_profile_names: C{bool}
        @param return_profile_ids: return as list of profile IDs
        @type return_profile_ids: C{bool}
        @param return_session_names: return as list of X2Go session names
        @type return_session_names: C{bool}

        @return: a session list (as UUID hashes, objects, profile names/IDs or session names)
        @rtype: C{list}

        """
        return self._sessionsWithState('terminated', return_objects=return_objects, return_profile_names=return_profile_names, return_profile_ids=return_profile_ids, return_session_names=return_session_names)

    @property
    def has_running_sessions(self):
        """\
        Equals C{True} if the underlying L{X2GoClient} instance has any running sessions at hand.

        """
        return self.running_sessions() and len(self.running_sessions()) > 0

    @property
    def has_suspended_sessions(self):
        """\
        Equals C{True} if the underlying L{X2GoClient} instance has any suspended sessions at hand.

        """
        return self.suspended_sessions and len(self.suspended_sessions) > 0

    def registered_sessions(self, return_objects=True, return_profile_names=False, return_profile_ids=False, return_session_names=False):
        """\
        Retrieve a list of all registered sessions.
        If none of the C{return_*} options is specified a list of session UUID hashes will be returned.

        @param return_objects: return as list of L{X2GoSession} instances
        @type return_objects: C{bool}
        @param return_profile_names: return as list of profile names
        @type return_profile_names: C{bool}
        @param return_profile_ids: return as list of profile IDs
        @type return_profile_ids: C{bool}
        @param return_session_names: return as list of X2Go session names
        @type return_session_names: C{bool}

        @return: a session list (as UUID hashes, objects, profile names/IDs or session names)
        @rtype: C{list}

        """
        return self._sessionsWithState('registered', return_objects=return_objects, return_profile_names=return_profile_names, return_profile_ids=return_profile_ids, return_session_names=return_session_names)

    def non_running_sessions(self, return_objects=True, return_profile_names=False, return_profile_ids=False, return_session_names=False):
        """\
        Retrieve a list of sessions that are currently _NOT_ in running state.
        If none of the C{return_*} options is specified a list of session UUID hashes will be returned.

        @param return_objects: return as list of L{X2GoSession} instances
        @type return_objects: C{bool}
        @param return_profile_names: return as list of profile names
        @type return_profile_names: C{bool}
        @param return_profile_ids: return as list of profile IDs
        @type return_profile_ids: C{bool}
        @param return_session_names: return as list of X2Go session names
        @type return_session_names: C{bool}

        @return: a session list (as UUID hashes, objects, profile names/IDs or session names)
        @rtype: C{list}

        """
        return [ s for s in self.registered_sessions(return_objects=return_objects, return_profile_names=return_profile_names, return_profile_ids=return_profile_ids, return_session_names=return_session_names) if s not in self.running_sessions(return_objects=return_objects, return_profile_names=return_profile_names, return_profile_ids=return_profile_ids, return_session_names=return_session_names) ]

    def connected_sessions_of_profile_name(self, profile_name, return_objects=True, return_session_names=False):
        """\
        For a given session profile name retrieve a list of sessions that are currently connected to the profile's X2Go server.
        If none of the C{return_*} options is specified a list of session UUID hashes will be returned.

        @param profile_name: session profile name
        @type profile_name: C{str}
        @param return_objects: return as list of L{X2GoSession} instances
        @type return_objects: C{bool}
        @param return_session_names: return as list of X2Go session names
        @type return_session_names: C{bool}

        @return: a session list (as UUID hashes, objects or session names)
        @rtype: C{list}

        """
        if return_objects:
            return self.connected_sessions() and [ s for s in self.connected_sessions() if s.get_profile_name() == profile_name ]
        elif return_session_names:
            return self.connected_sessions() and [ s.session_name for s in self.connected_sessions() if s.get_profile_name() == profile_name ]
        else:
            return self.connected_sessions() and [ s.get_uuid() for s in self.connected_sessions() if s.get_profile_name() == profile_name ]

    def associated_sessions_of_profile_name(self, profile_name, return_objects=True, return_session_names=False):
        """\
        For a given session profile name retrieve a list of sessions that are currently associated by an C{X2GoTerminalSession*} to this L{X2GoClient} instance.
        If none of the C{return_*} options is specified a list of session UUID hashes will be returned.

        @param profile_name: session profile name
        @type profile_name: C{str}
        @param return_objects: return as list of L{X2GoSession} instances
        @type return_objects: C{bool}
        @param return_session_names: return as list of X2Go session names
        @type return_session_names: C{bool}

        @return: a session list (as UUID hashes, objects or session names)
        @rtype: C{list}

        """
        if return_objects:
            return self.associated_sessions() and [ s for s in self.associated_sessions() if s.get_profile_name() == profile_name ]
        elif return_session_names:
            return self.associated_sessions() and [ s.session_name for s in self.associated_sessions() if s.get_profile_name() == profile_name ]
        else:
            return self.associated_sessions() and [ s.get_uuid() for s in self.associated_sessions() if s.get_profile_name() == profile_name ]

    def pubapp_sessions_of_profile_name(self, profile_name, return_objects=True, return_session_names=False):
        """\
        For a given session profile name retrieve a list of sessions that can be providers for published application list.
        If none of the C{return_*} options is specified a list of session UUID hashes will be returned.

        @param profile_name: session profile name
        @type profile_name: C{str}
        @param return_objects: return as list of L{X2GoSession} instances
        @type return_objects: C{bool}
        @param return_session_names: return as list of X2Go session names
        @type return_session_names: C{bool}

        @return: a session list (as UUID hashes, objects or session names)
        @rtype: C{list}

        """
        if return_objects:
            return self.associated_sessions_of_profile_name(profile_name) and [ s for s in self.associated_sessions_of_profile_name(profile_name) if s.is_published_applications_provider() ]
        elif return_session_names:
            return self.associated_sessions_of_profile_name(profile_name) and [ s.session_name for s in self.associated_sessions_of_profile_name(profile_name) if s.is_published_applications_provider() ]
        else:
            return self.associated_sessions_of_profile_name(profile_name) and [ s.get_uuid() for s in self.associated_sessions_of_profile_name(profile_name) if s.is_published_applications_provider() ]

    def registered_sessions_of_profile_name(self, profile_name, return_objects=True, return_session_names=False):
        """\
        For a given session profile name retrieve a list of sessions that are currently registered with this L{X2GoClient} instance.
        If none of the C{return_*} options is specified a list of session UUID hashes will be returned.

        @param profile_name: session profile name
        @type profile_name: C{str}
        @param return_objects: return as list of L{X2GoSession} instances
        @type return_objects: C{bool}
        @param return_session_names: return as list of X2Go session names
        @type return_session_names: C{bool}

        @return: a session list (as UUID hashes, objects or session names)
        @rtype: C{list}

        """
        if return_objects:
            return self.registered_sessions() and [ s for s in self.registered_sessions() if s.get_profile_name() == profile_name ]
        elif return_session_names:
            return self.registered_sessions() and [ s.session_name for s in self.registered_sessions() if s.get_profile_name() == profile_name ]
        else:
            return self.registered_sessions() and [ s.get_uuid() for s in self.registered_sessions() if s.get_profile_name() == profile_name ]

    def virgin_sessions_of_profile_name(self, profile_name, return_objects=True, return_session_names=False):
        """\
        For a given session profile name retrieve a list of sessions that are registered with this L{X2GoClient} instance but have not
        yet been started (i.e. sessions that are in virgin state). If none of the C{return_*} options is specified a list of 
        session UUID hashes will be returned.

        @param profile_name: session profile name
        @type profile_name: C{str}
        @param return_objects: return as list of L{X2GoSession} instances
        @type return_objects: C{bool}
        @param return_session_names: return as list of X2Go session names
        @type return_session_names: C{bool}

        @return: a session list (as UUID hashes, objects or session names)
        @rtype: C{list}

        """
        if return_objects:
            return self.virgin_sessions() and [ s for s in self.virgin_sessions() if s.get_profile_name() == profile_name ]
        elif return_session_names:
            return self.virgin_sessions() and [ s.session_name for s in self.virgin_sessions() if s.get_profile_name() == profile_name ]
        else:
            return self.virgin_sessions() and [ s.get_uuid() for s in self.virgin_sessions() if s.get_profile_name() == profile_name ]

    def running_sessions_of_profile_name(self, profile_name, return_objects=True, return_session_names=False):
        """\
        For a given session profile name retrieve a list of sessions that are currently running.
        If none of the C{return_*} options is specified a list of session UUID hashes will be returned.

        @param profile_name: session profile name
        @type profile_name: C{str}
        @param return_objects: return as list of L{X2GoSession} instances
        @type return_objects: C{bool}
        @param return_session_names: return as list of X2Go session names
        @type return_session_names: C{bool}

        @return: a session list (as UUID hashes, objects or session names)
        @rtype: C{list}

        """
        if return_objects:
            return self.running_sessions() and [ s for s in self.running_sessions() if s.get_profile_name() == profile_name ]
        elif return_session_names:
            return self.running_sessions() and [ s.session_name for s in self.running_sessions() if s.get_profile_name() == profile_name ]
        else:
            return self.running_sessions() and [ s.get_uuid() for s in self.running_sessions() if s.get_profile_name() == profile_name ]

    def suspended_sessions_of_profile_name(self, profile_name, return_objects=True, return_session_names=False):
        """\
        For a given session profile name retrieve a list of sessions that are currently in suspended state.
        If none of the C{return_*} options is specified a list of session UUID hashes will be returned.

        @param profile_name: session profile name
        @type profile_name: C{str}
        @param return_objects: return as list of L{X2GoSession} instances
        @type return_objects: C{bool}
        @param return_session_names: return as list of X2Go session names
        @type return_session_names: C{bool}

        @return: a session list (as UUID hashes, objects or session names)
        @rtype: C{list}

        """
        if return_objects:
            return self.suspended_sessions() and [ s for s in self.suspended_sessions() if s.get_profile_name() == profile_name ]
        elif return_session_names:
            return self.suspended_sessions() and [ s.session_name for s in self.suspended_sessions() if s.get_profile_name() == profile_name ]
        else:
            return self.suspended_sessions() and [ s.get_uuid() for s in self.suspended_sessions() if s.get_profile_name() == profile_name ]

    def control_session_of_profile_name(self, profile_name):
        """\
        For a given session profile name retrieve a the corresponding C{X2GoControlSession*} instance.

        @param profile_name: session profile name
        @type profile_name: C{str}

        @return: contol session instance
        @rtype: C{X2GoControlSession*} instance

        """
        _sessions = self.registered_sessions_of_profile_name(profile_name, return_objects=True)
        if _sessions:
            session = _sessions[0]
            return session.control_session
        return None

    @property
    def connected_control_sessions(self):
        """\
        Equals a list of all currently connected control sessions.

        """
        return [ c for c in self.control_sessions.values() if c.is_connected() ]

    def connected_profiles(self, use_paramiko=False, return_profile_ids=True, return_profile_names=False):
        """\
        Retrieve a list of all currently connected session profiles.

        @param use_paramiko: send query directly to the Paramiko/SSH layer
        @type use_paramiko: C{bool}

        @return: list of connected session profiles
        @rtype: C{list}

        """
        if use_paramiko:
            return [ p for p in self.control_sessions.keys() if self.control_sessions[p].is_connected() ]
        else:
            return self.connected_sessions(return_profile_ids=return_profile_ids, return_profile_names=return_profile_names)

    def get_master_session(self, profile_name, return_object=True, return_session_name=False):
        """\
        Retrieve the master session of a specific profile.

        @param profile_name: the profile name that we query the master session of
        @type profile_name: C{str}
        @param return_object: return L{X2GoSession} instance
        @type return_object: C{bool}
        @param return_session_name: return X2Go session name
        @type return_session_name: C{bool}

        @return: a session list (as UUID hashes, objects, profile names/IDs or session names)
        @rtype: C{list}

        """
        if profile_name not in self.connected_profiles(return_profile_names=True):
            return None

        if profile_name not in self.master_sessions.keys() or self.master_sessions[profile_name] is None:
            return None

        _session = self.master_sessions[profile_name]

        if not _session.is_master_session():
            del self.master_sessions[profile_name]
            return None

        if return_object:
            return _session
        elif return_session_name:
            return _session.get_session_name()
        else:
            return _session.get_uuid()



# -*- coding: utf-8 -*-

# Copyright (C) 2010-2013 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
For MIME box jobs there are currently three handling actions available:
L{X2GoMIMEboxActionOPEN}, L{X2GoMIMEboxActionOPENWITH} and L{X2GoMIMEboxActionSAVEAS}.

"""
__NAME__ = 'x2gomimeboxactions-pylib'

# modules
import os
import copy
import time

from defaults import X2GOCLIENT_OS as _X2GOCLIENT_OS
if _X2GOCLIENT_OS in ("Windows"):
    import subprocess
    import win32api
else:
    import gevent_subprocess as subprocess
    import x2go_exceptions
    WindowsErrror = x2go_exceptions.WindowsError

# Python X2Go modules
import log
import x2go_exceptions

_MIMEBOX_ENV = os.environ.copy()


class X2GoMIMEboxAction(object):

    __name__ = 'NAME'
    __description__ = 'DESCRIPTION'

    def __init__(self, client_instance=None, logger=None, loglevel=log.loglevel_DEFAULT):
        """\
        This is a meta class and has no functionality as such. It is used as parent
        class by »real« X2Go MIME box actions.

        @param client_instance: the underlying L{X2GoClient} instance
        @type client_instance: C{obj}
        @param logger: you can pass an L{X2GoLogger} object to the
            L{X2GoMIMEboxAction} constructor
        @type logger: C{obj}
        @param loglevel: if no L{X2GoLogger} object has been supplied a new one will be
            constructed with the given loglevel
        @type loglevel: C{int}

        """
        if logger is None:
            self.logger = log.X2GoLogger(loglevel=loglevel)
        else:
            self.logger = copy.deepcopy(logger)
        self.logger.tag = __NAME__

        # these get set from within the X2GoMIMEboxQueue class
        self.profile_name = 'UNKNOWN'
        self.session_name = 'UNKNOWN'

        self.client_instance = client_instance

    @property
    def name(self):
        """\
        Return the X2Go MIME box action's name.

        """
        return self.__name__

    @property
    def description(self):
        """\
        Return the X2Go MIME box action's description text.

        """
        return self.__description__

    def _do_process(self, mimebox_file, mimebox_dir, ):
        """\
        Perform the defined MIME box action (doing nothing in L{X2GoMIMEboxAction} parent class).

        @param mimebox_file: file name as placed in to the X2Go MIME box directory
        @type mimebox_file: C{str}
        @param mimebox_dir: location of the X2Go session's MIME box directory
        @type mimebox_dir: C{str}

        """
        pass

    def do_process(self, mimebox_file, mimebox_dir, ):
        """\
        Wrapper method for the actual processing of MIME
        box actions.

        @param mimebox_file: file name as placed in to the X2Go MIME box directory
        @type mimebox_file: C{str}
        @param mimebox_dir: location of the X2Go session's MIME box directory
        @type mimebox_dir: C{str}

        """
        mimebox_file = os.path.normpath(mimebox_file)
        mimebox_dir = os.path.normpath(mimebox_dir)

        self._do_process(mimebox_file, mimebox_dir)


class X2GoMIMEboxActionOPEN(X2GoMIMEboxAction):
    """\
    MIME box action that opens incoming files in the system's default application.

    """
    __name__= 'OPEN'
    __decription__= 'Open incoming file with local system\'s default application.'

    def __init__(self, client_instance=None, logger=None, loglevel=log.loglevel_DEFAULT):
        """\
        @param client_instance: the underlying L{X2GoClient} instance
        @type client_instance: C{obj}
        @param logger: you can pass an L{X2GoLogger} object to the
            L{X2GoMIMEboxActionOPEN} constructor
        @type logger: C{obj}
        @param loglevel: if no L{X2GoLogger} object has been supplied a new one will be
            constructed with the given loglevel
        @type loglevel: C{int}

        """
        self.client_instance = client_instance
        X2GoMIMEboxAction.__init__(self, logger=logger, loglevel=loglevel)

    def _do_process(self, mimebox_file, mimebox_dir, ):
        """\
        Open an incoming MIME box file in the system's default application.

        @param mimebox_file: file name as placed in to the MIME box directory
        @type mimebox_file: C{str}
        @param mimebox_dir: location of the X2Go session's MIME box directory
        @type mimebox_dir: C{str}

        """
        mimebox_file = os.path.normpath(mimebox_file)
        mimebox_dir = os.path.normpath(mimebox_dir)

        if _X2GOCLIENT_OS == "Windows":
            self.logger('opening incoming MIME box file with Python\'s os.startfile() command: %s' % mimebox_file, loglevel=log.loglevel_DEBUG)
            try:
                os.startfile(os.path.join(mimebox_dir, mimebox_file))
            except WindowsError, win_err:
                if self.client_instance:
                    self.client_instance.HOOK_mimeboxaction_error(mimebox_file,
                                                                  profile_name=self.profile_name,
                                                                  session_name=self.session_name,
                                                                  err_msg=str(win_err)
                                                                 )
                else:
                    self.logger('Encountered WindowsError: %s' % str(win_err), loglevel=log.loglevel_ERROR)
            time.sleep(20)
        else:
            cmd_line = [ 'xdg-open', os.path.join(mimebox_dir, mimebox_file), ]
            self.logger('opening MIME box file with command: %s' % ' '.join(cmd_line), loglevel=log.loglevel_DEBUG)
            subprocess.Popen(cmd_line, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=_MIMEBOX_ENV)
            time.sleep(20)


class X2GoMIMEboxActionOPENWITH(X2GoMIMEboxAction):
    """\
    MIME box action that calls the system's ,,Open with...'' dialog on incoming files. Currently only
    properly implementable on Windows platforms.

    """
    __name__= 'OPENWITH'
    __decription__= 'Evoke ,,Open with...\'\' dialog on incoming MIME box files.'

    def __init__(self, client_instance=None, logger=None, loglevel=log.loglevel_DEFAULT):
        """\
        @param client_instance: the underlying L{X2GoClient} instance
        @type client_instance: C{obj}
        @param logger: you can pass an L{X2GoLogger} object to the
            L{X2GoMIMEboxActionOPENWITH} constructor
        @type logger: C{obj}
        @param loglevel: if no L{X2GoLogger} object has been supplied a new one will be
            constructed with the given loglevel
        @type loglevel: C{int}

        """
        self.client_instance = client_instance
        X2GoMIMEboxAction.__init__(self, logger=logger, loglevel=loglevel)

    def _do_process(self, mimebox_file, mimebox_dir, ):
        """\
        Open an incoming MIME box file in the system's default application.

        @param mimebox_file: file name as placed in to the MIME box directory
        @type mimebox_file: C{str}
        @param mimebox_dir: location of the X2Go session's MIME box directory
        @type mimebox_dir: C{str}

        """
        mimebox_file = os.path.normpath(mimebox_file)
        mimebox_dir = os.path.normpath(mimebox_dir)

        if _X2GOCLIENT_OS == "Windows":
            self.logger('evoking Open-with dialog on incoming MIME box file: %s' % mimebox_file, loglevel=log.loglevel_DEBUG)
            win32api.ShellExecute (
                  0,
                  "open",
                  "rundll32.exe",
                  "shell32.dll,OpenAs_RunDLL %s" % os.path.join(mimebox_dir, mimebox_file),
                  None,
                  0,
            )
            time.sleep(20)
        else:
            self.logger('the evocation of the Open-with dialog box is currently not available on Linux, falling back to MIME box action OPEN', loglevel=log.loglevel_WARN)
            cmd_line = [ 'xdg-open', os.path.join(mimebox_dir, mimebox_file), ]
            self.logger('opening MIME box file with command: %s' % ' '.join(cmd_line), loglevel=log.loglevel_DEBUG)
            subprocess.Popen(cmd_line, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=_MIMEBOX_ENV)
            time.sleep(20)


class X2GoMIMEboxActionSAVEAS(X2GoMIMEboxAction):
    """\
    MIME box action that allows saving incoming MIME box files to a local folder. What this 
    MIME box actually does is calling a hook method in the L{X2GoClient} instance that
    can be hi-jacked by one of your application's methods which then can handle the ,,Save as...''
    request.

    """
    __name__ = 'SAVEAS'
    __decription__= 'Save incoming file as...'

    def __init__(self, client_instance=None, logger=None, loglevel=log.loglevel_DEFAULT):
        """\
        @param client_instance: an L{X2GoClient} instance, within your customized L{X2GoClient} make sure 
            you have a C{HOOK_open_mimebox_saveas_dialog(filename=<str>)} method defined that will actually
            handle the incoming mimebox file.
        @type client_instance: C{obj}
        @param logger: you can pass an L{X2GoLogger} object to the
            L{X2GoMIMEboxActionSAVEAS} constructor
        @type logger: C{obj}
        @param loglevel: if no L{X2GoLogger} object has been supplied a new one will be
            constructed with the given loglevel
        @type loglevel: C{int}

        @raise X2GoMIMEboxActionException: if the client_instance has not been passed to the SAVEAS MIME box action

        """
        if client_instance is None:
            raise x2go_exceptions.X2GoMIMEboxActionException('the SAVEAS MIME box action needs to know the X2GoClient instance (client=<instance>)')
        X2GoMIMEboxAction.__init__(self, client_instance=client_instance, logger=logger, loglevel=loglevel)

    def _do_process(self, mimebox_file, mimebox_dir):
        """\
        Call an L{X2GoClient} hook method (C{HOOK_open_mimebox_saveas_dialog}) that
        can handle the MIME box's SAVEAS action.

        @param mimebox_file: file name as placed in to the MIME box directory
        @type mimebox_file: C{str}
        @param mimebox_dir: location of the X2Go session's MIME box directory
        @type mimebox_dir: C{str}
        @param mimebox_file: PDF file name as placed in to the X2Go spool directory

        """
        mimebox_file = os.path.normpath(mimebox_file)
        mimebox_dir = os.path.normpath(mimebox_dir)

        self.logger('Session %s (%s) is calling X2GoClient class hook method <client_instance>.HOOK_open_mimebox_saveas_dialog(%s)' % (self.session_name, self.profile_name, mimebox_file), loglevel=log.loglevel_NOTICE)
        self.client_instance.HOOK_open_mimebox_saveas_dialog(os.path.join(mimebox_dir, mimebox_file), profile_name=self.profile_name, session_name=self.session_name)
        time.sleep(60)




# -*- coding: utf-8 -*-

# Copyright (C) 2010-2013 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
X2GoClientSettings class - managing x2goclient settings file (incl. LDAP-Support).

The L{X2GoClientSettings} class one of Python X2Go's a public API classes.
Use this class (indirectly by retrieving it from an L{X2GoClient} instance)
in your Python X2Go based applications to access the
»settings« configuration file of your X2Go client application.

"""
__NAME__ = 'x2gosettings-pylib'

# Python X2Go modules
import x2go.log as log
from x2go.defaults import X2GO_SETTINGS_CONFIGFILES as _X2GO_SETTINGS_CONFIGFILES
from x2go.defaults import X2GO_CLIENTSETTINGS_DEFAULTS as _X2GO_CLIENTSETTINGS_DEFAULTS
import x2go.inifiles as inifiles


class X2GoClientSettingsFILE(inifiles.X2GoIniFile):
    """\
    Configuration file based settings for L{X2GoClient} instances.

    """
    defaultValues = _X2GO_CLIENTSETTINGS_DEFAULTS

    def __init__(self, config_files=_X2GO_SETTINGS_CONFIGFILES, defaults=None, logger=None, loglevel=log.loglevel_DEFAULT):
        """\
        Constructs an L{X2GoClientSettingsFILE} instance. This is normally done from within an L{X2GoClient} instance.
        You can retrieve this L{X2GoClientSettingsFILE} instance with the L{X2GoClient.get_client_settings()} 
        method.

        On construction the L{X2GoClientSettingsFILE} object is filled with values from the configuration files::

            /etc/x2goclient/settings
            ~/.x2goclient/settings

        The files are read in the specified order and config options of both files are merged. Options 
        set in the user configuration file (C{~/.x2goclient/settings}) override global options set in
        C{/etc/x2goclient/settings}.

        """
        inifiles.X2GoIniFile.__init__(self, config_files, defaults=defaults, logger=logger, loglevel=loglevel)



# -*- coding: utf-8 -*-

# Copyright (C) 2010 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
# 
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
This file is a default test runner as found in ZOPE/Plone products. It works 
fine for any kind of Python unit testing---as we do here for Python X2Go.
"""

import os
import sys

base = os.path.join(os.path.split(os.path.split(os.getcwd())[0])[0])

# prepend the X2Go path (useful for building new packages)
sys.path = [os.path.normpath(base)] + sys.path

import unittest
TestRunner = unittest.TextTestRunner
suite = unittest.TestSuite()

tests = os.listdir(os.curdir)
tests = [n[:-3] for n in tests if n.startswith('test') and n.endswith('.py')]

for test in tests:
    m = __import__(test)
    if hasattr(m, 'test_suite'):
        suite.addTest(m.test_suite())

if __name__ == '__main__':
    TestRunner().run(suite)



# -*- coding: utf-8 -*-

# Copyright (C) 2010-2013 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

"""\
L{X2GoSessionProfiles} class - managing x2goclient session profiles.

L{X2GoSessionProfiles} is a public API class. Use this class in your Python X2Go based 
applications.

"""
__NAME__ = 'x2gosessionprofiles-pylib'

import copy
import types
import re

# Python X2Go modules
from x2go.defaults import X2GO_SESSIONPROFILES_CONFIGFILES as _X2GO_SESSIONPROFILES_CONFIGFILES
from x2go.defaults import X2GO_SESSIONPROFILE_DEFAULTS as _X2GO_SESSIONPROFILE_DEFAULTS
from x2go.defaults import X2GO_DESKTOPSESSIONS as _X2GO_DESKTOPSESSIONS
import x2go.inifiles as inifiles
import x2go.log as log
import x2go.utils as utils
from x2go.x2go_exceptions import X2GoProfileException


class X2GoSessionProfilesFILE(inifiles.X2GoIniFile):

    defaultSessionProfile = _X2GO_SESSIONPROFILE_DEFAULTS
    _non_profile_sections = ('embedded')

    def __init__(self, config_files=_X2GO_SESSIONPROFILES_CONFIGFILES, defaults=None, session_profile_defaults=None, logger=None, loglevel=log.loglevel_DEFAULT):
        """\
        Retrieve X2Go session profiles from a file, typically C{~/.x2goclient/sessions}.

        @param config_files: a list of config file locations, the first file name in this list the user has write access to will be the user configuration file
        @type config_files: C{list}
        @param defaults: not used for this class
        @type defaults: C{dict}
        @param session_profile_defaults: a default session profile
        @type session_profile_defaults: C{dict}
        @param logger: you can pass an L{X2GoLogger} object to the
            L{X2GoSessionProfilesFILE} constructor
        @type logger: L{X2GoLogger} instance
        @param loglevel: if no L{X2GoLogger} object has been supplied a new one will be
            constructed with the given loglevel
        @type loglevel: C{int}

        """
        self.defaultValues = {}
        self._profile_metatypes = {}
        self._cached_profile_ids = []
        self._cached_profile_names = []
        self._profiles_need_profile_id_renewal = []

        if logger is None:
            self.logger = log.X2GoLogger(loglevel=loglevel)
        else:
            self.logger = copy.deepcopy(logger)
        self.logger.tag = __NAME__

        # providing defaults for an X2GoSessionProfiles instance will---in the worst case---override your
        # existing sessions file in your home directory once you write the sessions back to file...
        inifiles.X2GoIniFile.__init__(self, config_files, defaults=defaults, logger=logger, loglevel=loglevel)

        if utils._checkSessionProfileDefaults(session_profile_defaults):
            self.defaultSessionProfile = session_profile_defaults

        self.session_profiles = [ p for p in self.iniConfig.sections() if p not in self._non_profile_sections ]
        for session_profile in self.session_profiles:
            for key, default_value in self.defaultSessionProfile.iteritems():
                if not self.iniConfig.has_option(session_profile, key):
                    self._storeValue(session_profile, key, default_value)
            self.get_profile_metatype(session_profile)

    def __call__(self, profile_id_or_name):
        """\
        Retrieve the session profile configuration for a given session profile ID (or name)

        @param profile_id_or_name: profile ID or profile name
        @type profile_id_or_name: C{str}

        @return: the profile ID's / name's profile configuration
        @rtype: C{dict}

        """
        _profile_id = self.check_profile_id_or_name(self, profile_id_or_name)
        return self.get_profile_config(profile_id=_profile_id)

    def get_profile_metatype(self, profile_id_or_name, force=False):
        """\
        Detect a human readable session profile type from the session profile configuration.

        @param profile_id_or_name: profile ID or profile name
        @type profile_id_or_name: C{str}
        @param force: re-detect the meta type, otherwise use a cached result
        @type force: C{bool}

        @return: the profile ID's / name's meta type
        @rtype: C{str}

        """
        _profile_id = self.check_profile_id_or_name(profile_id_or_name)
        if not self._profile_metatypes.has_key(_profile_id) or force:
            _config = self.get_profile_config(_profile_id)
            if _config['host']:
                if _config['rdpserver'] and _config['command'] == 'RDP':
                    _metatype = 'RDP/proxy'
                elif _config['published']:

                    if _config['command'] in _X2GO_DESKTOPSESSIONS.keys():
                        _metatype = '%s + Published Applications' % _config['command']
                    else:
                        _metatype = 'Published Applications'

                elif _config['rootless']:
                    _metatype = 'Single Applications'
                elif _config['command'] in _X2GO_DESKTOPSESSIONS.keys():
                    _metatype = '%s Desktop' % _config['command']
                elif _config['command'] in _X2GO_DESKTOPSESSIONS.values():
                    _metatype = '%s Desktop' % [ s for s in _X2GO_DESKTOPSESSIONS.keys() if _config['command'] == _X2GO_DESKTOPSESSIONS[s] ][0]
                else:
                    _metatype = 'CUSTOM Desktop'
            else:
                if _config['rdpserver'] and _config['command'] == 'RDP':
                    _metatype = 'RDP/direct'
                else:
                    _metatype = 'not supported'
            self._profile_metatypes[_profile_id] = _metatype
        else:
            return self._profile_metatypes[_profile_id]

    def write(self):

        # then update profile IDs for profiles that have a renamed host attribute...
        for profile_id in self._profiles_need_profile_id_renewal:
            _config = self.get_profile_config(profile_id=profile_id)
            self.iniConfig.remove_section(profile_id)
            try: self._cached_profile_ids.remove(profile_id)
            except ValueError: pass
            self.add_profile(profile_id=None, force_add=True, **_config)
        self._profiles_need_profile_id_renewal = []
        self._cached_profile_ids = []
        self._cached_profile_names = []

        # at last write the profile config as is...
        return inifiles.X2GoIniFile.write(self)


    def get_profile_option_type(self, option):
        """\
        Get the data type for a specific session profile option.

        @param option: the option to get the data type for
        @type option: will be detected by this method

        @return: the data type of C{option}
        @rtype: C{type}

        """
        try:
            return type(self.defaultSessionProfile[option])
        except KeyError:
            return types.StringType

    def get_type(self, section, key):
        """\
        Override the parent class's get_type method due to the special layout of this class.

        @param section: INI file section
        @type section: C{str}
        @param key: key in INI file section
        @type key: C{str}

        @return: the data type of C{key} in C{section}
        @rtype: C{type}

        """
        # we have to handle the get_type method separately...
        return self.get_profile_option_type(key)

    def get_profile_config(self, profile_id_or_name=None, parameter=None, profile_id=None):
        """\
        The configuration options for a single session profile.

        @param profile_id_or_name: either profile ID or profile name is accepted
        @type profile_id_or_name: C{str}
        @param parameter: if specified, only the value for the given parameter is returned
        @type parameter: C{str}
        @param profile_id: profile ID (faster than specifying C{profile_id_or_name})
        @type profile_id: C{str}

        @return: the session profile configuration for the given profile ID (or name)
        @rtype: C{dict}

        """
        _profile_id = profile_id or self.check_profile_id_or_name(profile_id_or_name)
        _profile_config = {}
        for option in self.iniConfig.options(_profile_id):
            _profile_config[option] = self.get(_profile_id, option, key_type=self.get_profile_option_type(option))

        if parameter is not None:
            if parameter in _profile_config.keys():

                value = _profile_config[parameter]

                if parameter == 'export':
                    _strvalue = value.replace(',', ';').strip().strip('"').strip().strip(';').strip()
                    value = {}
                    if _strvalue:
                        _export_paths = _strvalue.split(';')
                        for _path in _export_paths:
                            if not re.match('.*:(0|1)$', _path): _path = '%s:1' % _path
                            _auto_export_path = re.match('.*:1$', _path) and True or False
                            _export_path = ':'.join(_path.split(':')[:-1])
                            value[_export_path] = _auto_export_path

                return value

            else:
                raise X2GoProfileException('no such session profile parameter: %s' % parameter)
        return _profile_config

    def default_profile_config(self):
        """\
        Return a default session profile.

        @return: default session profile
        @rtype: C{dict}

        """
        return copy.deepcopy(self.defaultSessionProfile)

    def has_profile(self, profile_id_or_name):
        """\
        Does a session profile of a given profile ID or profile name exist?

        @param profile_id_or_name: profile ID or profile name
        @type profile_id_or_name: C{str}

        @return: C{True} if there is such a session profile, C{False} otherwise
        @rtype: C{bool}

        """
        try:
            self.check_profile_id_or_name(profile_id_or_name)
            return True
        except X2GoProfileException:
            return False

    @property
    def profile_ids(self):
        """\
        Renders a list of all profile IDs found in the session profile configuration file.

        """
        if not self._cached_profile_ids:
            self._cached_profile_ids = [ s for s in self.iniConfig.sections() if s not in self._non_profile_sections ]
        return self._cached_profile_ids

    def has_profile_id(self, profile_id):
        """\
        Does a session profile of a given profile ID exist? (Faster than L{has_profile()}.)

        @param profile_id: profile ID
        @type profile_id: C{str}

        @return: C{True} if there is such a session profile, C{False} otherwise
        @rtype: C{bool}

        """
        return profile_id in self.profile_ids

    @property
    def profile_names(self):
        """\
        Renders a list of all profile names found in the session profile configuration file.

        """
        if not self._cached_profile_names:
            self._cached_profile_names = [ self.to_profile_name(p) for p in self.profile_ids ]
        return  self._cached_profile_names

    def has_profile_name(self, profile_name):
        """\
        Does a session profile of a given profile name exist? (Faster than L{has_profile()}.)

        @param profile_name: profile name
        @type profile_name: C{str}

        @return: C{True} if there is such a session profile, C{False} otherwise
        @rtype: C{bool}

        """
        return profile_name in self.profile_names

    def to_profile_id(self, profile_name):
        """\
        Convert profile name to profile ID.

        @param profile_name: profile name
        @type profile_name: C{str}

        @return: profile ID
        @rtype: C{str}

        """
        _profile_ids = [ p for p in self.profile_ids if self.to_profile_name(p) == profile_name ]
        if len(_profile_ids) == 1:
            return _profile_ids[0]
        elif len(_profile_ids) == 0:
            return None
        else:
            raise X2GoProfileException('The sessions config file contains multiple session profiles with name: %s' % profile_name)

    def to_profile_name(self, profile_id):
        """\
        Convert profile ID to profile name.

        @param profile_id: profile ID
        @type profile_id: C{str}

        @return: profile name
        @rtype: C{str}

        """
        _profile_config = self.get_profile_config(profile_id=profile_id)
        if _profile_config.has_key('name'):
            return _profile_config['name']
        else:
            return ''

    def add_profile(self, profile_id=None, force_add=False, **kwargs):
        """\
        Add a new session profile.

        @param profile_id: a custom profile ID--if left empty a profile ID will be auto-generated
        @type profile_id: C{str}
        @param kwargs: session profile options for this new session profile
        @type kwargs: C{dict}

        @return: the (auto-generated) profile ID of the new session profile
        @rtype: C{str}

        """
        if profile_id is None or profile_id in self.profile_ids:
            profile_id = utils._genSessionProfileId()

        if 'name' not in kwargs.keys():
            raise X2GoProfileException('session profile parameter ,,name\'\' is missing in method parameters')

        if kwargs['name'] in self.profile_names and not force_add:
            raise X2GoProfileException('a profile of name ,,%s\'\' already exists' % kwargs['name'])

        for key, value in kwargs.items():
            self.update_value(None, key, value, profile_id=profile_id)

        for key, value in self.defaultSessionProfile.items():
            if key in kwargs: continue
            self.update_value(None, key, value, profile_id=profile_id)

        self._cached_profile_ids = []
        self._cached_profile_names = []

        return profile_id

    def delete_profile(self, profile_id_or_name):
        """\
        Delete a session profile from the configuration file.

        @param profile_id_or_name: profile ID or profile name
        @type profile_id_or_name: C{str}

        """
        _profile_id = self.check_profile_id_or_name(profile_id_or_name)
        self.iniConfig.remove_section(_profile_id)
        self.write_user_config = True
        self.write()
        self._cached_profile_ids = []
        self._cached_profile_names = []

    def update_value(self, section, key, value, profile_id=None):
        """\
        Update a value in a session profile.

        @param section: the profile ID
        @type section: C{str}
        @param key: the session profile option of the given profile ID
        @type key: C{str}
        @param value: the value to update the session profile option with
        @type value: any type, depends on the session profile option

        """
        try:
            profile_id = profile_id or self.check_profile_id_or_name(section)
        except X2GoProfileException:
            profile_id = section

        if key == 'name':
            profile_name = value
            current_profile_name = self.get_value(profile_id, key)
            if not profile_name:
                raise X2GoProfileException('profile name for profile id %s may not be empty' % profile_id)
            else:
                self._cached_profile_names = []
                if profile_name != current_profile_name and profile_name in self.profile_names:
                    raise X2GoProfileException('a profile of name ,,%s\'\' already exists' % profile_name)

        if key == 'export' and type(value) == types.DictType:
            _strvalue = '"'
            for folder in value.keys():
                _strvalue += "%s:%s;" % (folder, int(value[folder]))
            _strvalue += '"'
            _strvalue = _strvalue.replace('""', '')
            value = _strvalue

        if key == 'host':
            _config = self.get_profile_config(profile_id=profile_id)
            if _config.has_key('host') and _config['host'] != value:
                self._profiles_need_profile_id_renewal.append(profile_id)

        inifiles.X2GoIniFile.update_value(self, profile_id, key, value)

    def check_profile_id_or_name(self, profile_id_or_name):
        """\
        Detect the profile ID from a given string which maybe profile ID or profile name.

        @param profile_id_or_name: profile ID or profile name
        @type profile_id_or_name: C{str}

        @return: profile ID
        @rtype: C{str}

        @raise X2GoProfileException: if no such session profile exists

        """
        _profile_id = None
        if self.has_profile_name(profile_id_or_name):
            # we were given a sesion profile name...
            _profile_id = self.to_profile_id(profile_id_or_name)
        elif self.has_profile_id(profile_id_or_name):
            # we were given a session profile id...
            _profile_id = profile_id_or_name
        else:
            raise X2GoProfileException('No session profile with id or name ,,%s\'\' exists.' % profile_id_or_name)
        return _profile_id

    def to_session_params(self, profile_id_or_name=None, profile_id=None):
        """\
        Convert session profile options to L{X2GoSession} constructor method parameters.

        @param profile_id_or_name: either profile ID or profile name is accepted
        @type profile_id_or_name: C{str}
        @param profile_id: profile ID (fast than specifying C{profile_id_or_name})
        @type profile_id: C{str}

        @return: a dictionary of L{X2GoSession} constructor method parameters
        @rtype: C{dict}

        """
        _profile_id = profile_id or self.check_profile_id_or_name(profile_id_or_name)
        return utils._convert_SessionProfileOptions_2_SessionParams(self.get_profile_config(_profile_id))

    def get_session_param(self, profile_id_or_name, param):
        """\
        Get a single L{X2GoSession} parameter from a specific session profile.

        @param profile_id_or_name: either profile ID or profile name is accepted
        @type profile_id_or_name: C{str}
        @param param: the parameter name in the L{X2GoSession} constructor method
        @type param: C{str}

        @return: the value of the session profile option represented by C{param}
        @rtype: depends on the session profile option requested

        """
        return self.to_session_params(profile_id_or_name)[param]



# -*- coding: utf-8 -*-

# Copyright (C) 2010-2013 by Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
#
# Python X2Go is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Python X2Go is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program; if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.

from x2go.defaults import BACKEND_CONTROLSESSION_DEFAULT

from _stdout import X2GoControlSessionSTDOUT

X2GoControlSession = eval(BACKEND_CONTROLSESSION_DEFAULT)
